#!/usr/bin/env python3

"""
Test misc.
"""

import os, time
import numpy as np
from .. import api

# If SD tests cannot run then in most cases treat this as a warning only.
# There is not much point in letting dozens of tests fail for the same
# reason. Once Seismic Store is fully supported, make one and only one
# test that fails if we cannot connect.
#
# Tests do not hard code credentials, so either check that the fallback
# environment has been set, or explicitly pick those up here.
try:
    import sdglue as sd
except Exception as ex:
    print("seismic store access via sdglue is not available:", ex)
    sd = None
if sd and not (os.getenv("OPENZGY_SDURL") and os.getenv("OPENZGY_SDAPIKEY")):
    print("seismic store access requires $OPENZGY_SDURL and $OPENZGY_SDAPIKEY")
    sd = None

def SDTestData(name):
    """
    Return the full path of a test file stored in Seismic Store.
    The default sd:// uri is hard coded but may be overridden by
    setting $OPENZGY_SDTESTDATA. The expected test data must have
    been uploaded to that location. The tests need read access and
    should preferably not have write access. If name is empty then
    just return the prefix without a trailing slash.
    """
    project = os.getenv("OPENZGY_SDTESTDATA", "sd://sntc/testdata")
    if project and project[-1] == "/": project = project[:-1]
    return project + "/" + name if name else project

def SDTestSink(name):
    """
    Return the full path to where in Seismic Store the specified test
    output should be stored. The default sd:// uri is hard coded but
    may be overridden by setting $OPENZGY_SDTESTSINK. The tests need
    read/write access. If name is empty then just return the prefix
    without a trailing slash.
    """
    project = os.getenv("OPENZGY_SDTESTSINK", "sd://sntc/testsink/d")
    if project and project[-1] == "/": project = project[:-1]
    return project + "/" + name if name else project

def SDCredentials(**kwargs):
    """
    Convenience to hard code credentials for testing. Returns a dict.
    Picking up sdurl/sdapikey from the environment is redundant since
    the library already does this as a fallback.
    """
    result = {
        "sdurl":    os.getenv("OPENZGY_SDURL", ""),
        "sdapikey": os.getenv("OPENZGY_SDAPIKEY", ""),
        "sdtoken":  os.getenv("OPENZGY_TOKEN", "FILE:carbon.slbapp.com")
    }
    for k, v in kwargs.items():
        result[k] = v
    return result

def HasSeismicStore():
    """
    Used to enable or disable unit tests based on what software is loaded.
    TODO-Low structure the tests better, to make it less likely to need this.
    """
    return not sd is None

def HasZFPCompression():
    """
    Used to enable or disable unit tests based on what software is loaded.
    TODO-Low structure the tests better, to make it less likely to need this.
    """
    return ("ZFP" in api.ZgyKnownCompressors() and
            "ZFP" in api.ZgyKnownDecompressors())

class TempFileAutoDelete:
    """
    Arrange for this explicitly named temporary file to be deleted when
    the instance goes out of scope. Works both for seismic store and locally.

    A missing file is considered an error. The test somehow did not manage
    to create the file. If the test wants to delete the file itself then
    it should call disarm() after the deletion.

    If an exception is already pending then a missing or undeletable file
    only causes a warning to be printed. The pending exception is presumably
    more important.

    There is a fairly harmless race condition between the creation of the
    instance and the actual creation of the file. If an exception is thrown
    in that interval then a message is printed about not being able to delete
    the file. You may alternatively pass armed=False to the constructor and
    then invoke arm() at the exact point you know the file exists. Good luck.
    If you get an exception when creating a file on the cloud, you might not
    know whether the file got created or not.

    The recommended naming convention is to use both the current time and
    a per-process random number as a prefix. This makes it simpler to
    clean up files that in spite of the aito delete got left behind.

    The class can also be used for files intended to be persistent, by
    ensuring the file gets deleted if any error happened while creating
    it. Call disarm() only when you are sure you want to keep it.
    At that point your file is no longer considered temporary.

    Init takes a "silent" parameter which is only intended for unit tests.
    """
    def __init__(self, name, iocontext = None, *, armed = True, silent = False):
        self._name = name
        self._iocontext = iocontext
        self._silent = silent
        self._armed = armed

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self._armed and self._name:
            self.remove(self._name, self._iocontext,
                        raise_on_error = False if type else True,
                        silent = self._silent)
            self._armed = False
            #print("AUTO-DELETE", self._name)
        #else:
        #    print("AUTO-DELETE was disarmed.")

    def disarm(self):
        """
        Tell the instance to not try to delete the file. Either because we
        deleted it ourselves, or because we for some reason decided to keep it,
        or we want to guard agaist exceptions during create by temporarily
        disarming and then arming the file.
        """
        #print("DISARM", self._name)
        self._armed = False

    def arm(self):
        """
        Tell the instance to try to delete the file on exit.
        """
        #print("ARM", self._name)
        self._armed = True

    @property
    def name(self):
        return self._name

    @staticmethod
    def remove(name, iocontext = None, raise_on_error = True, silent = False):
        """
        Delete a file locally or in seismic store.
        Not using openzgy.api.ZgyUtils.delete() because that method
        will always suppress "file not found" errors. Please Don't do this
        unless you are ok with the explicit sdapi dependency.
        """
        credentials = ("", "", "")
        if iocontext:
            try:
                credentials = (iocontext["sdurl"],
                               iocontext["sdapikey"],
                               iocontext["sdtoken"])
            except (TypeError, KeyError):
                credentials = (iocontext.sdurl,
                               iocontext.sdapikey,
                               iocontext.sdtoken)
        if name:
            try:
                if name and name[:5] == "sd://":
                    if sd is not None:
                        with sd.SdUtil(credentials) as u:
                            u.delete(name)
                else:
                    os.remove(name)
            except Exception as ex:
                if not silent:
                    print("WARNING: Cannot AUTO-DELETE", name, str(type(ex)), str(ex))
                if raise_on_error:
                    raise

    @staticmethod
    def _randomname():
        try:
            with open("/dev/urandom", "rb") as random:
                n = np.frombuffer(random.read(4), dtype=np.uint32)[0]
        except FileNotFoundError: # Probably on a windows box.
            n =random.randint(0, 0xffffffff)
        return "{0:08x}-".format(n)

class LocalFileAutoDelete(TempFileAutoDelete):
    """
    As TempFileAutoDelete, but explicitly requesting a local file with
    a random prefix in from of the name.
    """
    _prefix = "tmp-{0:08x}-".format(int(time.time()))
    _prefix = os.path.join(os.getenv("TESTRUNDIR", '.'), _prefix)
    def __init__(self, suffix, *, silent = False):
        name = self._prefix + self._randomname() + suffix
        super().__init__(name, None, silent=silent)

class CloudFileAutoDelete(TempFileAutoDelete):
    """
    As TempFileAutoDelete, but explicitly requesting a file on the
    seismic store with a random prefix in from of the name.
    """
    _prefix = SDTestSink(None)
    _prefix += "/tmp-{0:08x}-".format(int(time.time()))
    def __init__(self, suffix, iocontext,*, silent = False):
        name = self._prefix + self._randomname() + (suffix or "tmp.tmp")
        super().__init__(name, iocontext, silent=silent)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
Test the low level decimation algorithms.
"""

import numpy as np
from enum import Enum
from ..impl.lodalgo import decimate, decimate8, DecimationType
from ..impl.histogram import HistogramData

def _make8(brick, special = None):
    """
    Expand the input brick to double size, with each sample in the input
    ending up in a 2x2x2 neighborhod with similar values (in the range
    sample_in..saple_in+6m average sample_in+3. Then chop the result into
    8 bricks having the same size as the original.
    Note, if using int8 then make sure the input values are at most 121.
    """
    half = brick.shape
    full = (half[0]*2, half[1]*2, half[2]*2)
    tmp = np.zeros(full, dtype=brick.dtype)
    tmp[0::2, 0::2, 0::2] = brick
    tmp[0::2, 0::2, 1::2] = brick + 1
    tmp[0::2, 1::2, 0::2] = brick + 2
    tmp[0::2, 1::2, 1::2] = brick + 3
    tmp[1::2, 0::2, 0::2] = brick + 4
    tmp[1::2, 0::2, 1::2] = brick + 5
    tmp[1::2, 1::2, 0::2] = brick + 6
    tmp[1::2, 1::2, 1::2] = brick + 3 if special is None else special
    bricks = []
    bricks.append(tmp[:half[0], :half[1], :half[2]].copy())
    bricks.append(tmp[:half[0], :half[1], half[2]:].copy())
    bricks.append(tmp[:half[0], half[1]:, :half[2]].copy())
    bricks.append(tmp[:half[0], half[1]:, half[2]:].copy())
    bricks.append(tmp[half[0]:, :half[1], :half[2]].copy())
    bricks.append(tmp[half[0]:, :half[1], half[2]:].copy())
    bricks.append(tmp[half[0]:, half[1]:, :half[2]].copy())
    bricks.append(tmp[half[0]:, half[1]:, half[2]:].copy())
    return bricks

def testLodAlgorithmsWithData(b0):
    """
    This will be run once with int8 data and once with float32
    data including NaNs. For most of the algorithms the expected
    result will be the same. In the int8 case the NaN values are
    replaced with a sample equal to the average of the other
    samples in that 2x2x2 region.
    """
    is_float = not np.issubdtype(b0.dtype, np.integer)
    bricks = _make8(b0, np.nan if is_float else None)
    # Make a variant of the test data that has one zero in each 2x2x2
    # neighborhood instead of a NaN, and that has no other zeros.
    b0x = b0.copy()
    b0x[b0x <= 0] = 1
    bricks_with_0 = _make8(b0x, 0)
    histogram = HistogramData(dtype=b0.dtype)
    histogram.add(bricks)
    result_LowPass  = decimate8(bricks, DecimationType.LowPass)
    result_WeightedAverage = decimate8(bricks, DecimationType.WeightedAverage, histogram = histogram, defaultvalue = np.nan)
    result_Average  = decimate8(bricks, DecimationType.Average)
    result_Median   = decimate8(bricks, DecimationType.Median)
    result_Minimum  = decimate8(bricks, DecimationType.Minimum)
    result_Maximum  = decimate8(bricks, DecimationType.Maximum)
    result_Decimate = decimate8(bricks, DecimationType.Decimate)
    result_AllZero  = decimate8(bricks, DecimationType.AllZero)
    result_AvgNon0  = decimate8(bricks_with_0, DecimationType.AverageNon0)
    assert np.all(result_Average  == b0 + 3)
    assert np.all(result_Median   == b0 + 3)
    assert np.all(result_Minimum  == b0)
    assert np.all(result_Maximum  == b0 + 6)
    assert np.all(result_Decimate == b0)
    assert np.all(result_AllZero  == np.zeros_like(bricks[0]))
    assert np.all(result_AvgNon0  == b0x + 3)
    # Expected result for LowPass is too difficult to compute if
    # a random cube is input. For slowly varying data it ought to
    # be similar to the Decimate case. And for all-constant data
    # the output should be that same constant.
    # The same reasoning applies to WeightedAverage.

def testLodAlgorithms():
    bricksize = (64, 64, 64)
    b0 = (np.random.random(bricksize) * 201 - 100).astype(np.int32)
    #print(b0[0,0,:])
    testLodAlgorithmsWithData(b0.astype(np.int8))
    testLodAlgorithmsWithData(b0.astype(np.float32))

def testLowpassLodAlgorithm():
    bricksize = (64, 64, 64)
    b0 = np.arange(np.product(bricksize), dtype=np.float64).reshape(bricksize)
    #b0.fill(42)
    bricks = _make8(b0, None) # TODO-Test also test NaN handling?
    lowpass = decimate8(bricks, DecimationType.LowPass)
    delta = lowpass - b0
    error = np.sqrt(np.average(delta * delta))
    #print("input  ", bricks[0][0,0,:])
    #print("lowpass", lowpass[0,0,:])
    #print("error  ", error)
    assert error < 0.1

    b0.fill(42)
    for b in bricks: b.fill(42)
    lowpass = decimate8(bricks, DecimationType.LowPass)
    delta = lowpass - b0
    error = np.sqrt(np.average(delta * delta))
    #print("input  ", bricks[0][0,0,:])
    #print("lowpass", lowpass[0,0,:])
    #print("error  ", error)
    assert error < 0.0001

    brick = [0.5, 0.1, 0.01] + 23*[0] + [1] + 37*[0]
    brick = np.array(4*brick)
    brick = brick.reshape((2, 2, len(brick)//4))
    d = decimate(brick, DecimationType.LowPass)
    # This input is simple enough that it makes sense to just
    # eyeball it. Not adding any asserts here.
    #print(brick[0,0,:])
    #print(d[0,0,:])

def testSpecial(dtype=np.float32, last=0):
    """
    There is some overlap with testLodAlgorithmsWithData(), but this
    test also does the newly added WeightedAverage algorithm which is
    important for Petrel.
    """
    cast = np.dtype(dtype).type
    cube = np.zeros((6, 10, 20), dtype=dtype)
    nibble = np.array([250, 42, 42, -1, -1, -1, 0, last], dtype=dtype)
    cube[:2,:2,:2] = nibble.reshape((2,2,2))
    # I want a total of 1*250, 5*42, 100*-1, 1094*0
    cube[2,0,:3] = 42
    cube[3,:10,:10] = -1 # add 100
    cube[3,0,:3] = 0 # and remove 3
    histogram = HistogramData(range_hint=(-3,+252), dtype=dtype)
    if dtype == np.int16:
        # Need a 1:1 mapping of inregral input data to bins
        histogram._size = 65536
        histogram._bins = np.zeros(histogram._size, dtype=np.int64)
    histogram.add(cube)

    # Expected weighted average computed by hand.
    s = 250/1 + 2*42/5 - 3*1/100 + 2*0/1094
    w = 1/1 + 2/5 + 3/100 + 2/1094
    if not np.isfinite(last): w -= 1/1094
    expect_wavg = s / w

    # Expected lowpass was just extracted from a previous run.
    expect_lowpass = 141.996475

    expect = {
        DecimationType.LowPass:          expect_lowpass,
        DecimationType.WeightedAverage:  expect_wavg,
        DecimationType.Average:          np.nanmean(nibble),
        DecimationType.Median:           0,
        DecimationType.Minimum:          -1,
        DecimationType.Maximum:          max(250, last),
        DecimationType.MinMax:           None,
        DecimationType.Decimate:         250,
        DecimationType.DecimateSkipNaN:  None,
        DecimationType.DecimateRandom:   None,
        DecimationType.AllZero:          0,
        DecimationType.WhiteNoise:       None,
        DecimationType.MostFrequent:     None,
        DecimationType.MostFrequentNon0: None,
        DecimationType.AverageNon0:      None, # Set below
    }

    # expect0 is the output when all 8 inputs are zero.
    expect0 = { k: (None if v is None else 0) for k, v in expect.items() }
    expect0[DecimationType.AverageNon0] = 99 # defaultvalue because all 0
    expect[DecimationType.AverageNon0] = np.mean([250, 42, 42, -1, -1, -1])
    if last == np.inf:
        # last entry is now in upper part instead of lower part.
        # The two elements at the center are 42 and 0, so the median
        # is reported as the average of those two.
        expect[DecimationType.Median] = 21

    actual = {}
    actual0 = {}
    for mode in expect:
        if expect[mode] is not None:
            if mode == DecimationType.WeightedAverage:
                a = decimate(cube, mode, histogram=histogram, defaultvalue=0)
            elif mode == DecimationType.AverageNon0:
                a = decimate(cube, mode, defaultvalue = 99)
            else:
                a = decimate(cube, mode)
            actual[mode] = a[0,0,0]
            actual0[mode] = a[2,0,0]

    errors = 0
    for mode in expect:
        if expect[mode] is not None:
            if not np.isclose(cast(expect[mode]), cast(actual[mode])):
                print("Error in decimate {0}: expect {1} got {2}".format(
                    mode.name, cast(expect[mode]), cast(actual[mode])))
                errors += 1
            elif not np.isclose(cast(expect0[mode]), cast(actual0[mode])):
                print("Error in decimate#2 {0}: expect {1} got {2}".format(
                    mode.name, cast(expect0[mode]), cast(actual0[mode])))
                errors += 1
            #else:
            #    print("Decimation {0}: got {1} and {2}".format(
            #        mode.name, cast(actual[mode]), cast(actual0[mode])))
    assert not errors

def numpyBugs():
    """
    I am fairly sure this is a bug in numpy. In operations on masked
    arrays a value of np.inf is sometimes treated as masked and sometimes
    not. The behavior is repeatable but changing an expression in a very
    minor fashion may change the result. Bottom line, try to avoid code
    that depends on this "feature".

    The unit test contains asserts cagainst the currently observed
    behavior in numpy 1.18.2. If the test starts failing after a numpy
    upgrade then this might mean that the bug has been fixed.
    """

    data1d = np.array([250, 42, 42, -1, -1, -1, 0, np.inf])
    data1d = np.ma.masked_equal(data1d, 0)
    data2d = data1d.reshape(-1,1)

    out1d = np.mean(data1d)
    out2d = np.mean(data2d, 0)[0]
    weird = np.mean(data2d, 1)

    def typename(x): return type(x).__module__ + "." + type(x).__name__
    #print(out1d, typename(out1d))
    #print(out2d, typename(out2d))
    #print(weird)

    # mean of 1d array with both masked elements and +inf.
    # Expected and actual results are positive infinity.
    assert str(out1d) == "inf"

    # mean over a single dimension of a 2d array, where the
    # other dimenzion is 1. Effectively the exact same data
    # except the result will be an array of one element.
    # This is NOT the case. The resulting value is masked.
    assert str(out2d) == "--"

    # Trying to do a mean over the other dimension, giving
    # 8 "mean" values where the input for each of them is
    # just a single value. Expected result is the original
    # data. Observed result is that the np.inf has been
    # replaced with a masked value.
    assert str(weird) == "[250.0 42.0 42.0 -1.0 -1.0 -1.0 -- --]"

    # To add to the confusion, if the out1d array is still a
    # masked array but starts out with nothing masked then the
    # behavior of the first case will change. Now this will
    # also return a masked value instead of inf.
    data1d = np.array([250, 42, 42, -1, -1, -1, 2, np.inf])
    data1d = np.ma.masked_equal(data1d, 0)
    out1d = np.mean(data1d)
    #print(out1d)
    assert str(out1d) == "--"

if __name__ == "__main__":
    np.seterr(all='raise')
    testLodAlgorithms()
    testLowpassLodAlgorithm()
    testSpecial(np.int16)
    testSpecial(np.float32)
    testSpecial(np.float32, np.nan)
    testSpecial(np.float32, np.inf)
    numpyBugs()

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



import numpy as np
import pandas as pd
import segyio

from ._segy_core import (
    _active_tracefield_segyio,
    _active_binfield_segyio,
    tqdm,
    check_tracefield,
)


TQDM_ARGS = dict(unit_scale=True, unit=" traces")


def segy_header_scan(segyfile, max_traces_scan=1000, silent=False, **segyio_kwargs):
    """Perform a scan of the segy file headers and return ranges.

    To get the raw header information see segy_header_scrape

    Args:
        segyfile (string): Segy File Path
        max_traces_scan (int/str, optional): Number of traces to scan.
            For all set to 0 or 'all'. Defaults to 1000.
        silent (bool): Disable progress bar.

    Returns:
        pandas.DataFrame: Uses pandas describe to return statistics of your headers.
    """
    if max_traces_scan == 0 or max_traces_scan == "all":
        max_traces_scan = None
    else:
        if not isinstance(max_traces_scan, int):
            raise ValueError("max_traces_scan must be int")

    head_df = segy_header_scrape(
        segyfile, max_traces_scan, silent=silent, **segyio_kwargs
    )

    header_keys = head_df.describe().T
    pre_cols = list(header_keys.columns)
    header_keys["byte_loc"] = list(_active_tracefield_segyio().values())
    header_keys = header_keys[["byte_loc"] + pre_cols]
    header_keys.nscan = head_df.shape[0]
    return header_keys


def segy_bin_scrape(segyfile, **segyio_kwargs):
    """Scrape binary header

    Args:
        segyfile (str): SEG-Y File path

    Returns:
        dict: Binary header
    """
    bk = _active_binfield_segyio()
    segyio_kwargs["ignore_geometry"] = True
    with segyio.open(segyfile, "r", **segyio_kwargs) as segyf:
        return {key: segyf.bin[item] for key, item in bk.items()}


def segy_header_scrape(
    segyfile,
    partial_scan=None,
    silent=False,
    bytes_filter=None,
    chunk=100_000,
    **segyio_kwargs,
):
    """Scape all data from segy trace headers

    Args:
        segyfile (str): SEG-Y File path
        partial_scan (int): Setting partial scan to a positive int will scan only
            that many traces. Defaults to None.
        silent (bool): Disable progress bar.
        bytes_filter (list): List of byte locations to load exclusively.
        chunk (int): Number of traces to read in one go.

    Returns:
        pandas.DataFrame: Raw header information in table for scanned traces.
    """
    check_tracefield(bytes_filter)

    assert (chunk > 0) and isinstance(chunk, int)
    header_keys = _active_tracefield_segyio()
    enum_byte_index = {
        int(byte_loc): i for i, byte_loc in enumerate(header_keys.values())
    }

    if bytes_filter:
        for byte_loc in bytes_filter:
            assert byte_loc in enum_byte_index
        bytes_filter_index = [enum_byte_index[byte_loc] for byte_loc in bytes_filter]
        enum_filter = [segyio.TraceField(byte_loc) for byte_loc in bytes_filter]
    else:
        bytes_filter_index = list(enum_byte_index.values())
        enum_filter = [segyio.TraceField(byte_loc) for byte_loc in header_keys.values()]

    columns = [list(header_keys.keys())[i] for i in bytes_filter_index]
    segyio_kwargs["ignore_geometry"] = True

    with segyio.open(segyfile, "r", **segyio_kwargs) as segyf:
        segyf_hgen = segyf.header[:]
        ntraces = segyf.tracecount
        if partial_scan is not None:
            ntraces = min(ntraces, int(partial_scan))

        head_df = pd.DataFrame(index=pd.Index(range(ntraces)), columns=columns)
        slc_end = chunk
        with tqdm(total=ntraces, disable=silent, **TQDM_ARGS) as pbar:
            while slc_end <= ntraces + chunk - 1:
                slc = slice(slc_end - chunk, min(slc_end, ntraces), 1)
                # take headers returned from segyio and create lists for a dataframe
                head_df.iloc[slc, :] = np.vstack(
                    [
                        list(next(segyf_hgen).values())
                        for _ in range(slc.stop - slc.start)
                    ]
                )[:, bytes_filter_index]
                slc_end += chunk
                pbar.update(slc.stop - slc.start)

    head_df.replace(to_replace=-2147483648, value=np.nan, inplace=True)
    return head_df


def what_geometry_am_i(head_df):
    """Try to work out file type from a header scan.

    This is a limited method to try and determine a file geometry type from
    a header scan. It will often fail.
    """

    ntraces = head_df.shape[0]

    stats = head_df.describe().T
    stats = stats[stats["std"] > 0]

    header_dict = _active_tracefield_segyio()
    pre_cols = list(stats.columns)
    stats["byte_loc"] = [
        header_dict[key] if key in header_dict.keys() else np.nan for key in stats.index
    ]
    stats = stats[["byte_loc"] + pre_cols]

    # find n unqiue
    stats["unique"] = [head_df[key].unique().size for key in stats.index]

    byte_call = dict()

    # check if offset
    if "offset" in stats.index:
        byte_call["offset"] = 37
        nof = stats.loc["offset", "unique"]
    else:
        nof = 1

    # check for xy
    if {"SourceX", "SourceY"}.issubset(set(stats.index)):
        byte_call["cdpx"] = 73
        byte_call["cdpy"] = 77
    elif {"CDP_X", "CDP_Y"}.issubset(set(stats.index)):
        has_xy = True
        byte_call["cdpx"] = 181
        byte_call["cdpy"] = 185

    # check for ilxl
    if {"INLINE_3D", "CROSSLINE_3D"}.issubset(set(stats.index)):
        byte_call["iline"] = 189
        byte_call["xline"] = 193
        n = stats.loc["INLINE_3D", "unique"] * stats.loc["CROSSLINE_3D", "unique"]
    elif "CDP" in stats.index:
        byte_call["cdp"] = 21
        n = stats.loc["CDP", "unique"]
    else:
        byte_call["cdp"] = 5
        n = stats["TRACE_SEQUENCE_FILE", "unique"]

    # sanity check number
    if nof * n != ntraces:
        raise ValueError(
            f"Couldn't determin geometry set byte locations manually. \n {print(stats)}"
        )

    # print(has_offsets)
    return byte_call



# pylint: disable=invalid-name,no-member
"""Xarray data accessor methods and functions to help with seismic analysis and
data manipulation.

"""
from collections.abc import Iterable

import xarray as xr
import numpy as np
import pandas as pd
from scipy.interpolate import griddata
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
from matplotlib.transforms import Affine2D

from ._keyfield import AttrKeyField, DimensionKeyField, CoordKeyField, VariableKeyField
from ._richstr import _upgrade_txt_richstr
from .tools import get_uniform_spacing, halfsample, plane


@xr.register_dataset_accessor("seisio")
class SeisIO:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def to_netcdf(self, seisnc, **kwargs):
        """Output to netcdf4 with specs for seisnc.

        Args:
            seisnc (string/path-like): The output file path. Preferably with .seisnc extension.
            **kwargs: As per xarray function to_netcdf.
        """
        # First remove all None attr.
        remove_keys = list()
        for key, val in self._obj.attrs.items():
            if val is None:
                remove_keys.append(key)
        for key in remove_keys:
            _ = self._obj.attrs.pop(key)

        # Convert richstring back to normal string
        try:
            self._obj.attrs["text"] = str(self._obj.attrs["text"])
        except:
            self._obj.attrs["text"] = ""

        kwargs["engine"] = "h5netcdf"

        self._obj.to_netcdf(seisnc, **kwargs)

    def to_subsurface(self):
        """Convert seismic data to a subsurface StructuredData Object

        Raises:
            err: [description]
            NotImplementedError: [description]

        Returns:
            subsurface.structs.base_structures.StructuredData: subsurface struct
        """
        try:
            from subsurface.structs.base_structures import StructuredData
        except ImportError as err:
            print("subsurface optional dependency required")
            raise err

        if self._obj.seis.is_twt():
            vdom = "twt"
        else:
            vdom = "depth"

        if self._obj.seis.is_3d():
            ds = self._obj.rename_dims({"iline": "x", "xline": "y", vdom: "z"})
            ds["z"] = ds["z"] * -1
            subsurface_struct = StructuredData(ds, "data")
        elif self._obj.seis.is_2d():
            subsurface_struct = StructuredData(self._obj, "data")
        else:
            raise NotImplementedError

        return subsurface_struct


def open_seisnc(seisnc, **kwargs):
    """Load from netcdf4 with seisnc specs.

    This all fills missing attributes required for output to other storage types.

    Args:
        seisnc (string/path-like): The input seisnc file.
        **kwargs: As per xarray function open_dataset.

    Returns:
        xarray dataset
    """
    kwargs["engine"] = "h5netcdf"
    ds = xr.open_dataset(seisnc, **kwargs)

    # Add back missing attr to remind people.
    for attr in AttrKeyField:
        if attr not in ds.attrs:
            ds.attrs[AttrKeyField[attr]] = None

    if ds.attrs["text"] is not None:
        ds.attrs["text"] = _upgrade_txt_richstr(ds.attrs["text"])

    promote_to_coord = [
        var for var in [CoordKeyField.cdp_x, CoordKeyField.cdp_y] if var in ds
    ]
    ds = ds.set_coords(promote_to_coord)

    return ds


def coordinate_df(seisnc, coord=True, extras=None, linear_fillna=True):
    """From an xarray seisnc quickly create iline and xline df. If coords is True
    also return cdp_x and cdp_y.

    Filling blank cdp_x and cdp_y spaces is particularly useful for using the
    xarray.plot option with x=cdp_x and y=cdp_y for geographic plotting.

    Args:
        seisnc (xarray.Dataset): The seisnc file.
        coord (bool, optional): Include cdp_x and cdp_y in putput.
        extras (list, optional): Any extra keywords to include in the dataframe
            from the seisnc variables.
        linear_fillna (bool, optional): Defaults to True. This will use a planar
            fitting function to fill blank cdp_x and cdp_y spaces.

    Returns:
        pandas.DataFrame

    """
    if coord:
        req_atr = ["iline", "xline", "cdp_x", "cdp_y"]
    else:
        req_atr = ["iline", "xline"]

    if extras is not None:
        req_atr += extras

    seisnc_coord = seisnc[req_atr].to_dataframe().reset_index()
    seisnc_coord_nona = seisnc_coord.dropna()

    if not linear_fillna:
        return seisnc_coord_nona

    # create fits to surface for target x from il/xl
    fitx = curve_fit(
        plane,
        (seisnc_coord_nona.iline, seisnc_coord_nona.xline),
        seisnc_coord_nona.cdp_x,
        p0=(0, 0, 0),
    )
    x_from_ix = lambda xy: plane(xy, *fitx[0])
    # create fits to surface for target y from il/xl
    fity = curve_fit(
        plane,
        (seisnc_coord_nona.iline, seisnc_coord_nona.xline),
        seisnc_coord_nona.cdp_y,
        p0=(0, 0, 0),
    )
    y_from_ix = lambda xy: plane(xy, *fity[0])

    filled_x = x_from_ix((seisnc_coord.iline, seisnc_coord.xline))
    filled_y = y_from_ix((seisnc_coord.iline, seisnc_coord.xline))

    seisnc_coord.loc[seisnc_coord.cdp_x.isna(), "cdp_x"] = filled_x[
        seisnc_coord.cdp_x.isna()
    ]
    seisnc_coord.loc[seisnc_coord.cdp_y.isna(), "cdp_y"] = filled_y[
        seisnc_coord.cdp_y.isna()
    ]

    return seisnc_coord


@xr.register_dataset_accessor("seis")
class SeisGeom:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    @property
    def humanbytes(self):
        """Prints Human Friendly size of Dataset to return the bytes as an
        int use ``xarray.Dataset.nbytes``

        Returns:
            str: Human readable size of dataset.

        """
        nbytes = self._obj.nbytes
        suffixes = ["B", "KB", "MB", "GB", "TB", "PB"]
        i = 0
        while nbytes >= 1024 and i < len(suffixes) - 1:
            nbytes /= 1024.0
            i += 1
        f = f"{nbytes:.2f}".rstrip("0").rstrip(".")
        return "{} {}".format(f, suffixes[i])

    def _coord_as_dimension(self, points, drop):
        """Convert x and y points to iline and xline. If the affine transform
        cannot be found this function will use a gridding interpolation approach.

        Args:
            cdp_x (float/array-like)
            cdp_y (float/array-like)
            method (str): Same a methods for xarray.Dataset.interp

        Returns:
            xarray.Dataset: At selected coordinates.
        """
        affine = None
        keys = ("cdp_x", "cdp_y")

        try:
            affine = self.get_affine_transform().inverted()
        except:
            pass

        if affine is not None:
            ilxl = affine.transform(np.dstack(points)[0])
            ils, xls = ilxl[:, 0], ilxl[:, 1]

        else:
            # check drop keys actually on dims, might not be
            drop = set(drop).intersection(*[set(self._obj[key].dims) for key in keys])

            grid = np.vstack(
                [
                    self._obj[key]
                    .mean(dim=drop, skipna=True)
                    .transpose("iline", "xline", transpose_coords=True)
                    .values.ravel()
                    for key in keys
                ]
            ).transpose()
            xlines_, ilines_ = np.meshgrid(
                self._obj["xline"],
                self._obj["iline"],
            )

            # these must all be the same length
            ils = np.atleast_1d(griddata(grid, ilines_.ravel(), points))
            xls = np.atleast_1d(griddata(grid, xlines_.ravel(), points))

        return ils, xls

    def xysel(self, cdp_x, cdp_y, method="nearest", sample_dim_name="cdp"):
        """Select data at x and y coordinates

        Args:
            cdp_x (float/array-like)
            cdp_y (float/array-like)
            method (str): Same as methods for xarray.Dataset.interp
            sample_dim_name (str, optional): The name to give the output sampling dimension.

        Returns:
            xarray.Dataset: At selected coordinates.
        """

        cdp_x = np.atleast_1d(cdp_x)
        cdp_y = np.atleast_1d(cdp_y)

        if self.is_twt():
            core_dims = DimensionKeyField.threed_twt
        elif self.is_depth():
            core_dims = DimensionKeyField.threed_depth
        else:
            raise AttributeError("Dataset required twt or depth coordinates.")

        dims = set(self._obj.dims.keys())

        if self.is_3d() or self.is_3dgath():
            # get other dims
            other_dims = dims.difference(core_dims)
            il, xl = self._coord_as_dimension((cdp_x, cdp_y), other_dims)

            sampling_arrays = dict(
                iline=xr.DataArray(
                    il, dims=sample_dim_name, coords={sample_dim_name: range(il.size)}
                ),
                xline=xr.DataArray(
                    xl, dims=sample_dim_name, coords={sample_dim_name: range(xl.size)}
                ),
            )
            output_ds = self._obj.interp(**sampling_arrays, method=method)

            # # populate back to 2d
            # cdp_ds = [
            #     None
            #     if np.isnan(np.sum([i, x]))
            #     else self._obj.interp(iline=i, xline=x, method=method)
            #     for i, x in zip(il, xl)
            # ]

            # if all(map(lambda x: x is None, cdp_ds)):
            #     raise ValueError("No points intersected the dataset")

            # for ds in cdp_ds:
            #     if ds is not None:
            #         none_replace = ds.copy(deep=True)
            #         break

            # none_replace[VariableKeyField.data][:] = np.nan
            # none_replace[CoordKeyField.iline] = np.nan
            # none_replace[CoordKeyField.xline] = np.nan

            # for i, (point, xloc, yloc) in enumerate(zip(cdp_ds, cdp_x, cdp_y)):
            #     if point is None:
            #         point = none_replace
            #     point[CoordKeyField.cdp_x] = point[CoordKeyField.cdp_x] * 0 + xloc
            #     point[CoordKeyField.cdp_y] = point[CoordKeyField.cdp_y] * 0 + yloc
            #     point[CoordKeyField.cdp] = i + 1
            #     point.attrs = dict()
            #     cdp_ds[i] = point.copy()

        elif self.is_2d() or self.is_2dgath():
            raise NotImplementedError("Not yet implemented for 2D")
        else:
            raise AttributeError(
                "xysel not support for this volume, must be 3d, 3dgath, 2d or 2dgath"
            )

        # cdp_ds = xr.concat(cdp_ds, "cdp")
        # cdp_ds.attrs = self._obj.attrs.copy()

        # return cdp_ds
        output_ds.attrs = self._obj.attrs.copy()
        return output_ds

    def _has_dims(self, dimension_options, invalid_dimension_options=None):
        """Check dataset has one of dimension options."""
        current_dims = set(self._obj.dims)

        # check for a match
        result = False
        for opt in dimension_options:
            if set(opt).issubset(current_dims):
                result = True
                break

        # check for things that would invalidate this e.g. this might be a subset
        if invalid_dimension_options is not None:
            for opt in invalid_dimension_options:
                if set(opt).issubset(current_dims):
                    result = False
                    break

        return result

    def _is_known_geometry(self):
        """Decide if known geometry this volume is and return the appropriate enum."""
        current_dims = set(self._obj.dims)
        for key in DimensionKeyField:
            if set(key).issubset(current_dims):
                return True
        return False

    def is_2d(self):
        """Returns True if the dataset is 2D peformant else False"""
        dim_options = [
            DimensionKeyField.twod_twt,
            DimensionKeyField.twod_depth,
        ]
        invalid_dim_options = [
            DimensionKeyField.twod_ps_twt,
            DimensionKeyField.twod_ps_depth,
        ]
        return self._has_dims(dim_options, invalid_dim_options)

    def is_3d(self):
        """Returns True if the dataset is 3D peformant else False"""
        dim_options = [
            DimensionKeyField.threed_twt,
            DimensionKeyField.threed_depth,
            DimensionKeyField.threed_xline_twt,
            DimensionKeyField.threed_iline_twt,
            DimensionKeyField.threed_xline_depth,
            DimensionKeyField.threed_iline_depth,
        ]
        invalid_dim_options = [
            DimensionKeyField.threed_ps_twt,
            DimensionKeyField.threed_ps_depth,
        ]
        return self._has_dims(dim_options, invalid_dim_options)

    def is_2dgath(self):
        """Returns True if the dataset is 2D peformant and has offset or angle else False"""
        dim_options = [
            DimensionKeyField.twod_ps_twt,
            DimensionKeyField.twod_ps_depth,
        ]
        return self._has_dims(dim_options)

    def is_3dgath(self):
        """Returns True if the dataset is 3D peformant and has offset or angle else False"""
        dim_options = [
            DimensionKeyField.threed_ps_twt,
            DimensionKeyField.threed_ps_depth,
        ]
        return self._has_dims(dim_options)

    def _check_multi_z(self):
        if (
            CoordKeyField.depth in self._obj.dims
            and CoordKeyField.twt in self._obj.dims
        ):
            raise ValueError(
                "seisnc cannot determine domain both twt and depth dimensions found"
            )

    def is_twt(self):
        """Check if seisnc volume is in twt"""
        self._check_multi_z()
        return True if CoordKeyField.twt in self._obj.dims else False

    def is_depth(self):
        """Check if seisnc volume is in depth"""
        self._check_multi_z()
        return True if CoordKeyField.depth in self._obj.dims else False

    def is_empty(self):
        """Check if empty"""
        if VariableKeyField.data in self._obj.variables:
            # This could be smartened up to check logic make sure dimensions of data
            # are correct.
            return False
        else:
            return True

    def get_measurement_system(self):
        """Return measurement_system if present, else None"""
        if hasattr(self._obj, "measurement_system"):
            return self._obj.measurement_system
        else:
            return None

    def zeros_like(self):
        """Create a new dataset with the same attributes and coordinates and
        dimensions but with data filled by zeros.
        """
        current_dims = {dim: val for dim, val in self._obj.dims.items()}
        dims = [dim for dim in current_dims.keys()]
        shp = [val for val in current_dims.values()]
        # TODO: Investigate copy options, might need to manually copy geometry
        # also don't want to load into memory if large large file.
        out = self._obj.copy()
        out[VariableKeyField.data] = (dims, np.zeros(shp))
        return out

    def surface_from_points(
        self,
        points,
        attr,
        left=("cdp_x", "cdp_y"),
        right=None,
        key=None,
        method="linear",
    ):
        """Sample a 2D point set with an attribute (like Z) to the seisnc
        geometry using interpolation.

        Args:
            points (array-like): Nx2 array-like of points with coordinates
                corresponding to coord1 and coord2.
            attr (array-like, str): If str points should be a DataFrame where
                attr is the column name of the attribute to be interpolated to
                the seisnc geometry. Else attr is a 1D array of length N.
            left (tuple): Length 2 tuple of coordinate dimensions to interpolate to.
            right (tuple, optional): If points is DataFrame right is a length 2
                tuple of column keys corresponding to coordinates in argument left.
            method (str): The interpolation method to use for griddata from scipy.
                Defaults to 'linear'

        Returns:
            xr.Dataset: Surface with geometry specified in left.

        """

        if len(left) != 2 and not isinstance(left, Iterable):
            raise ValueError("left must be tuple of length 2")
        for var in left:
            if var not in self._obj.variables:
                raise ValueError("left coordinate names not found in dataset")
        for var in left:
            if var in self._obj.dims:
                raise ValueError(
                    "surface_from_points is designed for coordinates that "
                    "are not already Dataset Dimensions, use ... instead."
                )

        a, b = left
        if self._obj[a].dims != self._obj[b].dims:
            raise ValueError(
                f"left keys {a} and {b} should exist on the same dimensions"
            )

        if isinstance(points, pd.DataFrame):
            if len(right) != 2 and not isinstance(right, Iterable):
                raise ValueError(
                    "points is DataFrame, right must a tuple specified "
                    "with coordinate keys from column names, e.g right=('cdpx', 'cdpy')"
                )
            names = [attr] + list(right)
            if not set(names).issubset(points.columns):
                raise ValueError("could not find specified keys in points DataFrame")

            _points = points[list(right)].values
            _attr = points[attr].values

        else:
            points = np.asarray(points)
            attr = np.asarray(attr)
            if points.shape[1] != 2:
                raise ValueError("points must have shape Nx2")
            if attr.size != points.shape[0]:
                raise ValueError("First dimension of points must equal length of attr")
            _points = points
            _attr = attr

        temp_ds = self._obj.copy()
        org_dims = temp_ds[a].dims
        org_shp = temp_ds[a].shape

        var = griddata(
            _points,
            _attr,
            np.dstack(
                [
                    temp_ds[a]
                    .transpose(*org_dims, transpose_coords=True)
                    .values.ravel(),
                    temp_ds[b]
                    .transpose(*org_dims, transpose_coords=True)
                    .values.ravel(),
                ]
            ),
            method=method,
        )

        out = temp_ds.drop_vars(temp_ds.var().keys())
        temp_ds.attrs.clear()
        if key is None and isinstance(attr, str):
            key = attr
        elif key is None:
            key = "data"
        out[key] = (org_dims, var.reshape(org_shp))
        return out

    # DRAFT
    # def get_survey_extents(seismic_3d):
    #     return SimpleNamespace(
    #         iline=(seismic_3d.coords["iline"].min().item(), seismic_3d.coords["iline"].max().item()),
    #         xline=(seismic_3d.coords["xline"].min().item(), seismic_3d.coords["xline"].max().item())
    #     )

    def calc_corner_points(self):
        """Calculate the corner points of the geometry or end points of a 2D line.

        This puts two properties in the seisnc attrs with the calculated il/xl and
        cdp_x and cdp_y if available.

        Attr:
            ds.attrs['corner_points']
            ds.attrs['corner_points_xy']
        """
        corner_points = False
        corner_points_xy = False

        if self.is_3d() or self.is_3dgath():
            il, xl = DimensionKeyField.cdp_3d
            ilines = self._obj[il].values
            xlines = self._obj[xl].values
            corner_points = (
                (ilines[0], xlines[0]),
                (ilines[0], xlines[-1]),
                (ilines[-1], xlines[-1]),
                (ilines[-1], xlines[0]),
            )

            cdpx, cdpy = CoordKeyField.cdp_x, CoordKeyField.cdp_y

            if set((cdpx, cdpy)).issubset(self._obj.variables):
                xs = self._obj[cdpx].values
                ys = self._obj[cdpy].values
                corner_points_xy = (
                    (xs[0, 0], ys[0, 0]),
                    (xs[0, -1], ys[0, -1]),
                    (xs[-1, -1], ys[-1, -1]),
                    (xs[-1, 0], ys[-1, 0]),
                )

        elif self.is_2d() or self.is_2dgath():
            cdp = DimensionKeyField.cdp_2d[0]
            cdps = self._obj[cdp].values
            corner_points = (cdps[0], cdps[-1])

            cdpx, cdpy = CoordKeyField.cdp_x, CoordKeyField.cdp_y

            if set((cdpx, cdpy)).issubset(self._obj.variables):
                xs = self._obj[cdpx].values
                ys = self._obj[cdpy].values
                corner_points_xy = (
                    (xs[0], ys[0]),
                    (xs[0], ys[0]),
                    (xs[-1], ys[1]),
                    (xs[-1], ys[1]),
                )

        if corner_points:
            self._obj.attrs[AttrKeyField.corner_points] = corner_points
        if corner_points_xy:
            self._obj.attrs[AttrKeyField.corner_points_xy] = corner_points_xy

    def interp_line(
        self,
        cdpx,
        cdpy,
        extra=None,
        bin_spacing_hint=10,
        line_method="slinear",
        xysel_method="linear",
    ):
        """Select data at x and y coordinates

        Args:
            cdp_x (float/array-like)
            cdp_y (float/array-like)
            bin_spacing_hint (number): a bin spacing to stay close to, in cdp world units. Default: 10
            line_method (string): valid values for the *kind* argument in scipy.interpolate.interp1d
            xysel_method (string): valid values for DataArray.interp

        Returns:
            xarray.Dataset: Interpolated traces along the arbitrary line
        """
        extra = dict() if extra is None else extra

        cdp_x_i, cdp_y_i, _, extra_i = get_uniform_spacing(
            cdpx,
            cdpy,
            bin_spacing_hint=bin_spacing_hint,
            extra=extra,
            method=line_method,
        )
        ds = self._obj.seis.xysel(cdp_x_i, cdp_y_i, method=xysel_method)

        for key in extra:
            ds[key] = (("cdp"), extra_i[key])

        return ds

    def plot_bounds(self, ax=None):
        """Plot survey bbounding box to a new or existing axis

        Args:
            ax: (optional) axis to plot to

        Returns:
            matplotlib axis used

        """
        xy = self._obj.attrs["corner_points_xy"]

        if xy is None:
            self._obj.seis.calc_corner_points()
            xy = self._obj.attrs["corner_points_xy"]

        if not ax:
            fig, ax = plt.subplots(1, 1)

        x = [x for x, _ in xy]
        y = [y for _, y in xy]
        ax.scatter(x, y)
        ax.plot(x + [x[0]], y + [y[0]], "k:")

        ax.set_xlabel("cdp x")
        ax.set_ylabel("cdp y")

        return ax

    def subsample_dims(self, **dim_kwargs):
        """Return a dictionary of subsampled dims suitable for xarray.interp.

        This tool halves

        Args:
            dim_kwargs: dimension names as keyword arguments with values of how
                many times we should divide the dimension by 2.
        """
        output = dict()
        for dim in dim_kwargs:
            ar = self._obj[dim].values
            while dim_kwargs[dim] > 0:  # only for 3.8
                ar = halfsample(ar)
                dim_kwargs[dim] = dim_kwargs[dim] - 1
            output[dim] = ar
        return output

    def fill_cdpna(self):
        """Fills NaN cdp locations by fitting known cdp x and y values to the
        local grid using a planar surface relationshipt.
        """
        coord_df = coordinate_df(self._obj)
        self._obj["cdp_x"] = (
            ("iline", "xline"),
            coord_df.cdp_x.values.reshape(self._obj.cdp_x.shape),
        )
        self._obj["cdp_y"] = (
            ("iline", "xline"),
            coord_df.cdp_y.values.reshape(self._obj.cdp_y.shape),
        )

    def get_affine_transform(self):
        """Calculate the forward iline/xline -> cdp_x, cdp_y Affine transform
        for Matplotlib using corner point geometry.

        Returns:
            matplotlib.transforms.Affine2D:

        Raises:
            ValueError: If Dataset is not 3D
        """
        if not self.is_3d():
            raise ValueError()
        self.calc_corner_points()

        cp_xy = self._obj.corner_points_xy
        cp_ix = self._obj.corner_points

        ang_xl = np.arctan2(cp_xy[1][1] - cp_xy[0][1], cp_xy[1][0] - cp_xy[0][0])
        ang_il = np.arctan2(cp_xy[3][1] - cp_xy[0][1], cp_xy[3][0] - cp_xy[0][0])
        il_length = np.sqrt(
            np.power(cp_xy[3][0] - cp_xy[0][0], 2)
            + np.power(cp_xy[3][1] - cp_xy[0][1], 2)
        )
        xl_length = np.sqrt(
            np.power(cp_xy[1][0] - cp_xy[0][0], 2)
            + np.power(cp_xy[1][1] - cp_xy[0][1], 2)
        )
        dxl = cp_ix[1][1] - cp_ix[0][1]
        dil = cp_ix[3][0] - cp_ix[0][0]

        affine_grid2loc = Affine2D()
        affine_grid2loc = (
            affine_grid2loc.rotate_around(  # rotate around origin
                cp_xy[0][0], cp_xy[0][1], -ang_il
            )
            .translate(-cp_xy[0][0], -cp_xy[0][1])  # move to zero origin
            .scale(1.0 / il_length, 1.0 / xl_length)  # scale to 1.0
            .scale(dil, dxl)  # scale to il/xl
            .translate(cp_ix[0][0], cp_ix[0][1])  # move to ilxl origin
        )
        return affine_grid2loc.inverted()

    def get_dead_trace_map(self, scan=None, zeros_as_nan=False):
        """Scan the vertical axis of a volume to find traces that are all NaN
        and return an DataArray which maps the all dead traces.

        Faster scans can be performed by setting scan to an int or list of int
        representing horizontal slice indexes to use for the scan.

        Args:
            scan (int/list of int, optional): Horizontal indexes to scan.
                Defaults to None (scan full volume).
            zeros_as_nan (bool, optional): Treat zeros as NaN during scan.

        Returns:
            xarray.DataArray: boolean dead trace map.
        """

        if self.is_twt():
            vdom = CoordKeyField.twt
        else:
            vdom = CoordKeyField.depth

        if scan:
            nan_map = (
                self._obj.data.isel(**{vdom: scan}).isnull().reduce(np.all, dim=vdom)
            )
        else:
            nan_map = self._obj.data.isnull().reduce(np.all, dim=vdom)

        if zeros_as_nan:
            zero_map = np.abs(self._obj.data).sum(dim=vdom)
            nan_map = xr.where(zero_map == 0.0, 1, nan_map)

        return nan_map



#!/usr/bin/env python3

"""
White-box unit tests for OpenZGY.
"""

import numpy as np
import os
import io
import sys
import math

from ..impl import bulk as impl_bulk
from ..impl import meta as impl_meta
from ..impl import file as impl_file
from ..impl import stats as impl_stats
from ..impl import enum as impl_enum
from ..impl import histogram as impl_histogram
from .. import api as openzgy
from ..exception import *
from ..test.utils import SDCredentials, LocalFileAutoDelete

def test_ScaleToStorage():
    """
    Unit test for ZgyInternalBulk._scaleToStorage.
    """
    fn = impl_bulk.ZgyInternalBulk._scaleToStorage
    codingrange = (-10.0, +30.0)

    # Converting the min- and max of the coding range should yield
    # the first and last valid storage number.
    assert fn(np.float32(-10.0), codingrange, np.int8) == -128
    assert fn(np.float32(+30.0), codingrange, np.int8) == 127
    assert fn(np.float32(-10.0), codingrange, np.uint16) == 0
    assert fn(np.float32(+30.0), codingrange, np.uint16) == 65535

    # Outside valid range should silently clip.
    assert fn(np.float32(-15.0), codingrange, np.int8) == -128
    assert fn(np.float32(+35.0), codingrange, np.int8) == 127

    # Conversion should round to nearest. For this coding range
    # each slot covers a bit more than +/- 0.0783
    assert fn(np.float32(-10.0 + 0.0783), codingrange, np.int8) == -128
    assert fn(np.float32(-10.0 + 0.0786), codingrange, np.int8) == -127
    assert fn(np.float32(-10.0 + 3*0.0783), codingrange, np.int8) == -127
    assert fn(np.float32(-10.0 + 3*0.0786), codingrange, np.int8) == -126

def test_LookupTables():
    """
    Unit test for the following related functions:
        _get{Brick,Alpha}FilePosition()
        _get{Brick,Alpha}LookupIndex()
        _validatePosition()
            lod out of range
            ijk out of range for lod 0
            ijk out of range for lod N, would not have been if lod == 0

    Calls these functions to get mocked data; these are assumed tested
    elsewhere. Or that they are so trivial that they don't need testing.
    """

    #test all-constant brick (both types) and missing brick
    # Arrange

    class MockInternalMeta: pass
    class MockInfoHeader: pass

    mock = MockInternalMeta()
    mock._ih = MockInfoHeader()
    mock._ih._size = (112, 64, 176) # bricks: (2, 1, 3)
    mock._ih._bricksize = (64, 64, 64)
    mock._ih._lodsizes = impl_meta._CalcLodSizes(mock._ih._size, mock._ih._bricksize)
    mock._ih._nlods = len(mock._ih._lodsizes)
    mock._ih._brickoffsets = impl_meta._CalcLutOffsets(mock._ih._lodsizes, False)
    mock._ih._alphaoffsets = impl_meta._CalcLutOffsets(mock._ih._lodsizes, True)
    mock._blup = [0] * mock._ih._brickoffsets[mock._ih._nlods]
    mock._alup = [0] * mock._ih._alphaoffsets[mock._ih._nlods]

    assert mock._ih._nlods == 3
    assert type(mock._ih._lodsizes) == tuple
    assert type(mock._ih._lodsizes[0]) == tuple
    assert tuple(mock._ih._lodsizes[0]) == (2, 1, 3)
    assert tuple(mock._ih._lodsizes[1]) == (1, 1, 2)
    assert tuple(mock._ih._lodsizes[2]) == (1, 1, 1)

def test_throws(e, f):
    try:
        f(None)
        return False
    except e:
        return True

def test_MapEnums():
    """
    Unit test for conversion between various internal and api enums.
    """
    from ..impl import enum as internal

    # (internal) RawDataType to (public) SampleDataType
    # Errors returned as unknown because the code should
    # be tolerant with regards to garbage in the file.
    assert openzgy._map_DataTypeToSampleDataType(internal.RawDataType.SignedInt8) == openzgy.SampleDataType.int8
    assert openzgy._map_DataTypeToSampleDataType(internal.RawDataType.SignedInt16) == openzgy.SampleDataType.int16
    assert openzgy._map_DataTypeToSampleDataType(internal.RawDataType.Float32) == openzgy.SampleDataType.float
    # Maps to unknown because I have not added this one to the enum
    assert openzgy._map_DataTypeToSampleDataType(internal.RawDataType.UnsignedInt8) == openzgy.SampleDataType.unknown

    # (public) SampleDataType to (internal) RawDataType
    # Errors should raise exceptions.
    assert openzgy._map_SampleDataTypeToDataType(openzgy.SampleDataType.int8) == internal.RawDataType.SignedInt8
    assert openzgy._map_SampleDataTypeToDataType(openzgy.SampleDataType.int16) == internal.RawDataType.SignedInt16
    assert openzgy._map_SampleDataTypeToDataType(openzgy.SampleDataType.float) == internal.RawDataType.Float32

    # (public) UnitDimension to (internal) Horizontal- and Vertical dim
    # Errors should raise exceptions.
    assert openzgy._map_UnitDimensionToHorizontalDimension(openzgy.UnitDimension.length) == internal.RawHorizontalDimension.Length
    assert openzgy._map_UnitDimensionToHorizontalDimension(openzgy.UnitDimension.arcangle) == internal.RawHorizontalDimension.ArcAngle
    # time is not allowed horizontally.
    assert test_throws(openzgy.ZgyUserError, lambda x: openzgy._map_UnitDimensionToHorizontalDimension(openzgy.UnitDimension.time))
    assert openzgy._map_UnitDimensionToVerticalDimension(openzgy.UnitDimension.length) == internal.RawVerticalDimension.Depth
    assert openzgy._map_UnitDimensionToVerticalDimension(openzgy.UnitDimension.time) == internal.RawVerticalDimension.SeismicTWT
    # arcangle not allowed vertically.
    assert test_throws(openzgy.ZgyUserError, lambda x: openzgy._map_UnitDimensionToVerticalDimension(openzgy.UnitDimension.arcangle))

    # (internal) Horizontal- and Vertical dim to (public) UnitDimension
    # Errors should return unknown, but if there was a bad tag it should
    # have been mapped to unknown already.
    assert openzgy._map_VerticalDimensionToUnitDimension(internal.RawVerticalDimension.SeismicTWT) == openzgy.UnitDimension.time
    assert openzgy._map_VerticalDimensionToUnitDimension(internal.RawVerticalDimension.SeismicOWT) == openzgy.UnitDimension.time
    assert openzgy._map_VerticalDimensionToUnitDimension(internal.RawVerticalDimension.Depth) == openzgy.UnitDimension.length
    assert openzgy._map_VerticalDimensionToUnitDimension(internal.RawVerticalDimension.Unknown) == openzgy.UnitDimension.unknown

    assert openzgy._map_HorizontalDimensionToUnitDimension(internal.RawHorizontalDimension.Length) == openzgy.UnitDimension.length
    assert openzgy._map_HorizontalDimensionToUnitDimension(internal.RawHorizontalDimension.ArcAngle) == openzgy.UnitDimension.arcangle
    assert openzgy._map_HorizontalDimensionToUnitDimension(internal.RawHorizontalDimension.Unknown) == openzgy.UnitDimension.unknown

    del internal


def test_FormatDescription():
    for verbose in (0, 1, 2):
        with io.StringIO() as stream:
            impl_meta.checkAllFormats(verbose=verbose, file=stream)
            # The part of checkAllFormats that would dump the definitions
            # in C++ or HTML have been moved to openzgy/tools/cppmeta.py.
            # The checks and now not expected to output anything.
            if verbose == 0: assert len(stream.getvalue()) == 0
            if verbose >  0: assert len(stream.getvalue()) == 0
            # No test on the result, only see that it doesn't crash.
            #print("FORMATS TYPE", verbose, stream.getvalue(), "@END")

def test_Statistics():
    stats = impl_stats.StatisticData()
    # adding scalars
    # -NOT. StatisticData() changed to only accept numpy arrays.
    stats += np.array(42)
    stats += np.array(2)
    stats += np.array(np.inf)
    stats += np.array(-np.inf)
    stats += np.array(-np.nan)
    #print(repr(stats))
    assert stats._cnt == 2
    assert stats._inf == 3
    assert stats._sum == 42 + 2
    assert stats._ssq == 42*42 + 2*2
    assert stats._min == 2
    assert stats._max == 42

    # adding arrays
    a = np.array([42, 2, np.inf, -np.inf, np.nan])
    stats2 = impl_stats.StatisticData()
    stats2 += a
    assert stats == stats2

    # Multiply by integral factor
    stats *= 3
    stats2 += a
    stats2 += a
    assert stats == stats2

    # Add two instances
    stats = stats + 3 * (impl_stats.StatisticData() + a)
    stats2 = stats2 + stats2

    # Scaling
    a = np.array([1, 0, 2, 3, 42], dtype=np.int8)
    stats = impl_stats.StatisticData()
    stats2 = impl_stats.StatisticData()
    stats += a
    stats2 += np.array([3.14*x + 2.72 for x in a])
    # old signature: stats.scale(0, 1, 2.72, 3.14+2.72)
    stats.scale(3.14, 2.72)
    #print(repr(stats))
    #print(repr(stats2))
    assert stats._cnt == stats2._cnt
    assert stats._inf == stats2._inf
    assert math.isclose(stats._sum, stats2._sum)
    assert math.isclose(stats._ssq, stats2._ssq)
    assert math.isclose(stats._min, stats2._min)
    assert math.isclose(stats._max, stats2._max)


def alltiles(reader):
    for lod in range(reader.nlods):
        size = reader.brickcount[lod]
        for ii in range(size[0]):
            for jj in range(size[1]):
                yield ii, jj, lod

def allbricks(reader):
    for lod in range(reader.nlods):
        size = reader.brickcount[lod]
        for ii in range(size[0]):
            for jj in range(size[1]):
                for kk in range(size[2]):
                    yield ii, jj, kk, lod

def _typename(t):
    return t.__name__ if t.__module__ is None or t.__module__ == type.__module__ else t.__module__ + "." + t.__name__

def test_bricksize(filename):
    with openzgy.ZgyReader(filename, iocontext = SDCredentials()) as reader:
        if False:
            print("Opened file size", reader.size, "bricks", reader.brickcount)
            print(list(allbricks(reader)))
        for ii, jj, kk, lod in allbricks(reader):
            ix = reader._accessor._getBrickLookupIndex(ii, jj, kk, lod)
            raw_beg = reader._accessor._metadata._blup._lookup[ix]
            raw_end = reader._accessor._metadata._blup._lookend[ix]
            if False:
                print("begin: {0}({1}), end: {2}({3})".format(
                      _typename(type(raw_beg)), hex(raw_beg),
                      _typename(type(raw_end)), hex(raw_end)))
            beg, size = reader._accessor._getBrickBegAndSize(ix, sane = False)
            if False:
                print("ix {ix} lod {lod} brickpos ({ii},{jj},{kk}) beg 0x{beg:x} size 0x{size:x}".format(
                    ix=ix, lod=lod, ii=ii, jj=jj, kk=kk, beg=beg, size=size))
            # Size will be > 64^3 sometimes due to interleaved alpha tiles
            # and it will be very large for the last brick since we skipped
            # the test for EOF. For Salt2-v3.zgy, lowres alpha tiles exist
            # between 0x440000 and 0x480000 so the brick at 0x400000 will
            # appear twice as large as it ought to be.
            assert beg == 0 or size >= 64*64*64

        for ii, jj, lod in alltiles(reader):
            ix = reader._accessor._getAlphaLookupIndex(ii, jj, lod)
            raw_beg = reader._accessor._metadata._alup._lookup[ix]
            raw_end = reader._accessor._metadata._alup._lookend[ix]
            if False:
                print("begin: {0}({1}), end: {2}({3})".format(
                      _typename(type(raw_beg)), hex(raw_beg),
                      _typename(type(raw_end)), hex(raw_end)))
            beg, size = reader._accessor._getAlphaBegAndSize(ix, sane = False)
            if False:
                print("ix {ix} lod {lod} alphapos ({ii},{jj}) beg 0x{beg:x} size 0x{size:x}".format(
                    ix=ix, lod=lod, ii=ii, jj=jj, beg=beg, size=size))
            assert beg == 0 or size >= 64*64

def test_PaddingOutsideSurvey(filename):
    with openzgy.ZgyWriter(filename,
                           iocontext = SDCredentials(),
                           size = (7, 13, 17),
                           datatype = openzgy.SampleDataType.int16,
                           datarange = (-32768,+32767)) as writer:
        data = np.arange(1, 7*13*17+1, dtype=np.int16).reshape((7,13,17))
        data = np.pad(data, ((0,64-7),(0,64-13),(0,64-17)),
                             mode='constant', constant_values=-999)
        writer.write((0,0,0), data)
        (brickstatus, fileoffset, constvalue, bricksize) = (
            writer._accessor._getBrickFilePosition(0, 0, 0, 0))

    # With ZgyReader the padding area should be filled with a default value,
    # which is zero in this simple case where there is no scale/offset.
    # So the only nonzero data should be our real samples.
    with openzgy.ZgyReader(filename, iocontext = SDCredentials()) as reader:
        readback = np.full((64, 64, 64), 42, dtype=np.int16)
        reader.read((0,0,0), readback)
        assert np.count_nonzero(readback) == 7 * 13 * 17

    # Reading using basic file I/O to get the data that the compressor saw.
    with open(filename, "rb") as f:
        f.seek(fileoffset, 0)
        check = f.read(64*64*64*2)
        check = np.frombuffer(check, dtype=np.int16).reshape((64, 64, 64))

    samples_in_survey = 7 * 13 * 17
    expect_nonzero_samples = 8 * 16 * 20
    actual_nonzero_samples = np.count_nonzero(check)

    assert np.all(data[:7,:13,:17] == check[:7,:13,:17])
    assert brickstatus == impl_enum.BrickStatus.Normal
    assert fileoffset == 64*64*64*2 # Not really a problem if it fails
    assert actual_nonzero_samples == expect_nonzero_samples
    assert check[7,12,16] == data[6,12,16] # one past lower corner
    assert check[6,13,16] == data[6,12,16]
    assert check[6,12,17] == data[6,12,16]
    assert np.all(check[7,:13,:17] == check[6,:13,:17])
    assert np.all(check[:7,13,:17] == check[:7,12,:17])
    assert np.all(check[:7,14,:17] == check[:7,12,:17])
    assert np.all(check[:7,15,:17] == check[:7,12,:17])
    assert np.all(check[:7,:13,17] == check[:7,:13,16])
    assert np.all(check[:7,:13,18] == check[:7,:13,16])
    assert np.all(check[:7,:13,19] == check[:7,:13,16])

def test_histogram16():
    hist = impl_histogram.HistogramData(dtype=np.int16)
    assert hist.vv_range == (-32768, +32767)
    if hist.bins.shape[0] != 65536:
        # I might not have enabled the feature of using oversize histograms.
        # But I can fake it by resizing the new, empty histogram.
        hist.resize(65536)
        assert hist.vv_range == (-32768, +32767)
        assert hist.bins.shape[0] == 65536
    # Using only 4096 of the possible 65536 values
    hist.add(1000+np.arange(4096, dtype=np.int16))
    assert np.all(hist.bins[32768+1000:32768+1000+4096] == 1)
    # The first number, 1000, ended up in bin 32768+1000.
    assert np.isclose(1000, hist.binvalue(32768+1000))
    hist.resize(256)
    # Current implementation reserves slot 0 for zero-centric
    # adjustment which is not implemented yet. The factor will
    # end up as 17, so we will have 4096/17 = 240 bins with count
    # 17 and the 16 left over in the bin immediately after those.
    assert len(hist.bins) == 256
    assert hist.bins[0] == 0
    assert np.all(hist.bins[1:241] == 17)
    assert hist.bins[241] == 16
    assert np.all(hist.bins[242:] == 0)
    #print(hist._hmin, hist._hmax, hist._bins, sep="\n")
    # The first number, 1000, should have moved to bin 1.
    assert np.isclose(1000, hist.binvalue(1))

def test_histogram8():
    hist = impl_histogram.HistogramData(dtype=np.int8)
    assert hist.vv_range == (-128, +127)
    # Using only 90 of the possible 256 values
    hist.add(10+np.arange(80, dtype=np.int8))
    assert np.all(hist.bins[128+10:128+10+80] == 1)
    # The first number, 10, ended up in bin 128+10.
    assert np.isclose(10, hist.binvalue(128+10))
    hist.resize(32)
    # Current implementation reserves slot 0 for zero-centric
    # adjustment which is not implemented yet. The factor will
    # end up as 3, so we will have 80/3 = 26 bins with count
    # 3 and the 2 left over in the bin immediately after those.
    assert len(hist.bins) == 32
    assert hist.bins[0] == 0
    assert np.all(hist.bins[1:1+26] == 3)
    assert hist.bins[27] == 2
    assert np.all(hist.bins[28:] == 0)
    #print(hist._hmin, hist._hmax, hist._bins, sep="\n")
    # The first number, 1000, should have moved to bin 1.
    assert np.isclose(10, hist.binvalue(1))
    # Now resize it larger. All the bins end up clumped together
    # in the lower past; the resize will simply add more empty
    # bins at the end.
    hist.resize(1000)
    assert len(hist.bins) == 1000
    assert hist.bins[0] == 0
    assert np.all(hist.bins[1:1+26] == 3)
    assert hist.bins[27] == 2
    assert np.all(hist.bins[28:] == 0)
    #print(hist._hmin, hist._hmax, hist._bins, sep="\n")
    # The first number, 1000, should have moved to bin 1.
    assert np.isclose(10, hist.binvalue(1))

if __name__ == "__main__":
    np.seterr(all='raise')
    test_histogram16()
    test_histogram8()
    test_ScaleToStorage()
    test_LookupTables()
    test_MapEnums()
    test_FormatDescription()
    test_Statistics()
    # TODO-High re-enable, use test data reachable from the cloud.
    #test_bricksize("/home/paal/git/Salmon/UnitTestData/Salmon/UnitTest/Salt2-v3.zgy")
    #test_bricksize("/home/paal/git/Salmon/UnitTestData/Salmon/UnitTest/Salt2-32.zgy")
    with LocalFileAutoDelete("padding.zgy") as fn:
        test_PaddingOutsideSurvey(fn.name)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



from ._segy_headers import segy_header_scrape, segy_bin_scrape, segy_header_scan
from ._segy_text import get_segy_texthead, put_segy_texthead, create_default_texthead
from ._segy_loader import (
    segy_loader,
    well_known_byte_locs,
    segy_converter,
    segy_freeloader,
)
from ._segy_writer import ncdf2segy, segy_writer, output_byte_locs
from ._segy_freewriter import segy_freewriter



# pylint: disable=invalid-name
"""
All functions for handling and manipulating SEISNC files. This also include the
basic functions for creating a valid blank seisnc spec Dataset and file.

Author:
    Tony Hallam 2020
"""

import warnings

warnings.warn(
    "the seisnc module is depreciated for seismic_dataset", DeprecationWarning
)

import netCDF4

import numpy as np
import xarray as xr

SEINC_DIMS = ["i", "j", "k"]
SEISNC_COORDS = ["iline", "xline", "z"]
SEISNC_NONDATA_VAR = ["CDP_X", "CDP_Y", "CDP_TRACE"]
SEISNC_ATTR = [
    "ns",
    "sample_rate",
    "text",
]

DS_VALID_COORDS = ["iline", "xline", "depth", "twt", "cdp_x", "cdp_y", "lat", "lon"]

DS_DEFAULT_ATTR = dict(
    ns=None,
    sample_rate=None,
    text=None,
    measurement_system=None,
    epsg=None,
    corner_points=None,
    corner_points_xy=None,
    source_file=None,
    srd=None,
    datatype=None,
)


def create_empty_seisnc(ncfile, dims):
    """Creates an on disk blank seisnc (NetCDF) file.

    Args:
        ncfile: The output file path.
        dims (tuple): The (inline, xline, vert) dimensions of the seismic cube.
    """
    # calculate vert, inline and crossline ranges/meshgrids
    ni, nx, ns = dims

    ilines = np.arange(ni, dtype=int)
    xlines = np.arange(nx, dtype=int)
    vert = np.arange(ns, dtype=int)

    # use i j k here to match REM and because xarray orders dimensions at this
    # step based upon the alphanumeric name
    seisnc = xr.Dataset(
        coords={"iind": (["i"], ilines), "jind": (["j"], xlines), "kind": (["k"], vert)}
    )
    seisnc.to_netcdf(path=ncfile, mode="w")


def set_seisnc_dims(
    ncfile,
    first_sample=0,
    sample_rate=1,
    first_iline=1,
    iline_step=1,
    first_xline=1,
    xline_step=1,
    vert_domain="TWT",
    measurement_system=0,
):
    """

    Args:
        ncfile ([type]): [description]
        first_sample (int, optional): [description]. Defaults to 0.
        sample_rate (int, optional): [description]. Defaults to 1.
        first_iline (int, optional): [description]. Defaults to 0.
        iline_step (int, optional): [description]. Defaults to 1.
        first_xline (int, optional): [description]. Defaults to 0.
        xline_step (int, optional): [description]. Defaults to 1.
        vert_domain (str, optional): [description]. Defaults to 'TWT'.
        measurement_system(int/str, optional): Measurement system of
            coordinates, one of ['m', 'ft' 0]: Defaults to 0 for unknown.
    """

    if vert_domain == "TWT":
        units = "ms"
    elif vert_domain == "Z":
        units = "m"

    with xr.open_dataset(ncfile) as seisnc:
        dims = seisnc.dims

    # create dimension arrays
    ns = dims["k"]
    ni = dims["i"]
    nx = dims["j"]
    vert = np.arange(
        first_sample, first_sample + sample_rate * ns, sample_rate, dtype=int
    )
    ilines = np.arange(
        first_iline, first_iline + iline_step * ni, iline_step, dtype=int
    )
    xlines = np.arange(
        first_xline, first_xline + xline_step * nx, xline_step, dtype=int
    )

    # update seisnc on disk
    with netCDF4.Dataset(str(ncfile), mode="a") as seisnc:
        seisnc.createVariable(
            "CDP_X",
            "f4",
            (
                "i",
                "j",
            ),
        )
        seisnc.createVariable(
            "CDP_Y",
            "f4",
            (
                "i",
                "j",
            ),
        )
        seisnc.createVariable(
            "CDP_TRACE",
            "i4",
            (
                "i",
                "j",
            ),
        )
        seisnc.createVariable("data", "f4", ("i", "j", "k"))

        seisnc.createVariable("vert", "f4", ("k"))
        seisnc.createVariable("iline", "i4", ("i"))
        seisnc.createVariable("xline", "i4", ("j"))

        seisnc["data"].units = units
        seisnc["iline"][:] = ilines
        seisnc["xline"][:] = xlines
        seisnc["vert"][:] = vert
        seisnc.ns = ns
        seisnc.sample_rate = sample_rate
        seisnc.measurement_system = measurement_system

        seisnc.text = "Blank seisnc"



#!/usr/bin/env python3

"""
Detect corrupt traces by looking for any trace having a sample magnitudes.
that is more than 10 times the 99-percentile of all magnitudes. If any are
found, set the entire trace to zero. Might also have considered clipping
to the 99-percentile instead.

To avoid running out of memory the 99-percentile is computed from max
magnitude of each trace, not of each sample. I guess I could save even
more by removing the bottom 99% of the collected & sorted values, ever
so often. Note: It would be nice to include his functionality in test_crop.
"""

import numpy as np
import os, sys, math
from ..api import ZgyReader, ZgyWriter, ProgressWithDots
from ..test.utils import SDCredentials
from ..iterator import readall

def round_sig(x, sig=2):
    return 0 if not x else round(x, sig-1-int(math.floor(math.log10(abs(x)))))

def scan_open_file(r, progress):
    limits = []
    for datastart, datasize, data in readall(r, progress=progress):
        lim = np.amax(np.abs(data), axis=2)
        limits.extend(map(float, lim.flat))
    limits.sort()
    cutoff = limits[(len(limits)*99)//100]
    print("Trace count: {0:,d}, 99-percentile: {1:.6g}".format(
        len(limits), cutoff))
    return round_sig(cutoff * 10)

def copy_open_file(r, w, cutoff, progress):
    nuked = 0
    total = 0
    for datastart, datasize, data in readall(r, progress=progress):
        lim = np.amax(np.abs(data), axis=2)
        for ii in range(lim.shape[0]):
            for jj in range(lim.shape[1]):
                total += 1
                if lim[ii,jj] > cutoff:
                    nuked += 1
                    data[ii,jj,:] = 0
                    if False:
                        print("trace {0},{1} max {2:.6g}".format(
                            datastart[0]+ii, datastart[1]+jj, lim[ii,jj]))
        w.write(datastart, data)
    print("{0:,d} of {1:,d} traces spiked > {2:.6g} and were zeroed.".format(
        nuked, total, cutoff))

def copy(srcfilename, dstfilename):
    ProgressWithDots()(100, 100)
    with ZgyReader(srcfilename, iocontext = SDCredentials()) as r:
        with ZgyWriter(dstfilename, templatename=srcfilename,
                       iocontext = SDCredentials()) as w:
            cutoff = scan_open_file(r, progress=ProgressWithDots())
            copy_open_file(r, w, cutoff, progress=ProgressWithDots())
            w.finalize(progress=ProgressWithDots())

if __name__ == "__main__":
    np.seterr(all='raise')
    copy(sys.argv[1], sys.argv[2])

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

#print('Running' if __name__ == '__main__' else 'Importing', __file__)

import numpy as np
import sys
import os
import time
from PIL import Image

from .. import api as newzgy
try:
    from .. import zgypublic as oldzgy
except Exception:
    oldzgy = None
from ..exception import ZgyMissingFeature
from ..test.utils import SDCredentials, SDTestData
from .viewzgy import savePNG

def read_data_all_at_once(reader, lod, start, size):
    section = np.zeros(size, dtype=np.float32)
    reader.read(start, section, lod = lod)
    return section

def read_data_b_at_a_time(reader, lod, start, size):
    bs = np.array(reader.bricksize, dtype=np.int64)
    padsize = ((np.array(size, np.int64) + bs - 1) // bs) * bs
    brick = np.zeros(bs, dtype=np.float32)
    section = np.zeros(padsize, dtype=np.float32)
    for ii in range(0, size[0], bs[0]):
        for jj in range(0, size[1], bs[1]):
            for kk in range(0, size[2], bs[2]):
                reader.read((start[0]+ii, start[1]+jj, start[2]+kk),brick, lod = lod)
                section[ii:ii+bs[0], jj:jj+bs[1], kk:kk+bs[2]] = brick
    return section[:size[0],:size[1],:size[2]]

def timing_report(reader, lod, start, size, elapsed):
    bs = np.array(reader.bricksize if isinstance(reader, newzgy.ZgyReader) else (64, 64, 64), dtype=np.int64)
    padsize = ((np.array(size, np.int64) + bs - 1) // bs) * bs
    bandwidth = np.product(padsize) / elapsed # should I use size or padsize?
    bandwidth /= (1024*1024)
    print("Elapsed {0:6.2f} seconds, bandwidth {1:6.2f} MVoxel/s reading {2} lod {3} size {4} start {5}".format(elapsed, bandwidth, reader.datatype, lod, tuple(size), tuple(start)))

_zgycloud_inited = False

def init_zgycloud():
    global _zgycloud_inited
    if _zgycloud_inited or oldzgy is None: return
    import zgycloud
    zgy = oldzgy.zgy
    #print("Glue together error handling", flush=True)
    zgy.setErrorHooks(zgycloud.resetError, zgycloud.lastError)
    #print("Handle logging inside Python", flush=True)
    zgycloud.setLogger(lambda pri,x: print("log:", x, end='', flush=True), 0)
    # Modify this to run the tests in a different environment.
    if False:
        zgycloud.configure('sd://', '',
                           '@/etc/delfi/demo-sdapi-key',
                           '@/etc/delfi/demo-sdapi-url')
    #print("Get test token", flush=True)
    token = zgycloud.getTestToken("sd://")
    zgycloud.setToken("sd://", token, "stoken", True)
    zgycloud.enableRealCache(True, 1024)
    zgycloud.setSegmentSize(1024)
    _zgycloud_inited = True

def run(filename, *, lods = [0], direction = 0, slurp=True,
        readerfactory = newzgy.ZgyReader, outname = None, iocontext=None):
    """
    Read 64 traces or slices in the specified direction.
    Optionally save the first of these as PNG.
    """
    if iocontext is None and filename[:5] == "sd://":
        iocontext = SDCredentials()
    if False:
        print("Read", filename,
              ("64 inlines", "64 crosslines", "64 slices")[direction],
              "slurp" if slurp else "block at a time")
    allslices = []
    with readerfactory(filename, iocontext=iocontext) as reader:
        slicenumber = reader.size[direction] // 2
        #reader.dump()
        for lod in lods:
            step = 1<<lod
            start = [0, 0, 0]
            size = np.array(reader.size, dtype=np.int64) // (1 << lod)
            start[direction] = slicenumber >> lod
            size[direction] = 1
            if direction == 2:
                size[0] = min(size[0], 1024)
                size[1] = min(size[1], 1024)
            start = tuple(start)
            size = tuple(size)
            starttime = time.time()
            if slurp:
                section = read_data_all_at_once(reader, lod, start, size)
            else:
                section = read_data_b_at_a_time(reader, lod, start, size)
            #timing_report(reader, lod, start, size, time.time() - starttime)
            # Display only the first section or slice
            if outname:
                if direction == 0:
                    myslice = section[0,...]
                elif direction == 1:
                    myslice = section[:,0,:]
                else:
                    myslice = section[...,0]
                #savePNG(myslice, outname + "_lod" + str(lod) + ".png")
                allslices.append(myslice)

    if outname:
        s = allslices[0].shape
        w = np.sum([allslices[i].shape[0] for i in range(3)])
        combined = np.zeros((w, s[1]), dtype=allslices[0].dtype)
        combined[0:s[0],:] = allslices[0]
        ss = allslices[1].shape
        combined[s[0]:s[0]+ss[0], 0:ss[1]] = allslices[1]
        sss = allslices[2].shape
        combined[s[0]+ss[0]:s[0]+ss[0]+sss[0], 0:sss[1]] = allslices[2]
        combined = combined[::-1,::]
        savePNG(combined, outname + ".png")

def Rema1000(filelist):
    """
    Specific test for the Rema 1000 solution: What is the best case
    and worst case effect of using 4 MB block size regardless of how
    the data is layed out?
    The file list should be one completely sorted, one completely random

    In all cases, read one brick at a time, strictly ordered by
    inline slowest, crossline, vertical fastest
    Currently the cache only holds the last block read, but
    in these lab conditions that should be enough.

    Control: config aligned = 0. Access direction and sorted or not
    shouldn't have any effect. If it does then there is probably
    some caching that I am not aware of, and it will be more
    difficult to interpret the results. Test all combinations of
    unsorted/sorted and access directions.

    Test: config aligned = 4 MB. What I expect / hope is that for a
    sorted file the inline- and crossline access should be
    significantly better. With inline access possibly a bit better
    than crossline. Slice access and all access to scrambled files
    should all be similar (assuming the controls were all similar)
    and hopefully not very much worse than the control.

    The entire test suite should be run both for 8-bit and float.
    Primarily test on cloud, but perhaps also on-prem with limited
    bandwidth.
    """
    for alignment in [0, 4]:
        print("Force alignment:", alignment)
        for filename in filelist:
            for direction in (0, 1, 2):
                run(filename, direction=direction, slurp=False,
                    readerfactory=newzgy.ZgyReader,
                    iocontext=SDCredentials(aligned=alignment))
    # For comparison, best case for the fancy reader.
    run(filelist[0], direction=0, slurp=True, readerfactory=newzgy.ZgyReader)


def Main(args):

    if not args:
        args = ["../build/testdata/Empty-v3.zgy",
                "../build/testdata/Empty-v1.zgy",
                SDTestData("SyntFixed.zgy")]

    if any([name[:5] == "sd://" for name in args]):
        init_zgycloud()

    suffix = 1
    for filename in args:
        out = os.path.join(os.getenv("TESTRUNDIR", '.'), "new" + str(suffix))
        try:
            run(filename, lods=range(3), direction=0, slurp=True,
                readerfactory=newzgy.ZgyReader, outname = out)
        except ZgyMissingFeature as ex:
            print("{0}: {1}".format(filename, str(ex)))

        out = os.path.join(os.getenv("TESTRUNDIR", '.'), "old" + str(suffix))
        if oldzgy is not None:
            run(filename, lods=range(3), direction=0, slurp=True,
                readerfactory=oldzgy.ZgyReader, outname = out)

        suffix += 1

if __name__ == "__main__":
    np.seterr(all='raise')
    Main(sys.argv[1:])
    sys.exit(0)
    Rema1000([SDTestData("bigtestdata/synt-50gb-8bit-sorted.zgy"),
              SDTestData("bigtestdata/synt-50gb-8bit-randomized.zgy")])

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



"""Fucntions to help with pretty printing in IPython
"""


def _prettify(self, p, cycle):
    if cycle:
        return p.text(self)
    else:
        lines = self.splitlines()
        p.text("Text Header")
        for line in lines:
            p.text(line)
            p.break_()


def _htmlify(self):
    html = f"<h3 style='font-size: medium;'>Text Header</h3>"
    lines = self.split("\n")
    html += "<h3 style='font-size: small;'>"
    for line in lines:
        html += f"{line}<br/>"
    html += "</h3>"
    return html


def _upgrade_txt_richstr(text):
    return type(
        "rich_texthead", (str,), dict(_repr_pretty_=_prettify, _repr_html_=_htmlify)
    )(text)



#!/usr/bin/python
# -*- coding: utf8 -*-+
import re
from setuptools import setup
from setuptools import find_packages

# from segysak import __version__

# Get README and remove badges.
readme = open("README.rst").read()

install_requires = [
    "numpy",
    "pandas",
    "segyio",
    "xarray",
    "dask",
    "distributed",
    "tqdm",
    "scipy",
    "click",
    "h5netcdf",
    "setuptools_scm",
    "attrdict",
    "more-itertools",
]

notebook_deps = [
    "ipython",
    "ipykernel>=4.8.0",  # because https://github.com/ipython/ipykernel/issues/274 and https://github.com/ipython/ipykernel/issues/263
    "jupyter_client>=5.2.2",  # because https://github.com/jupyter/jupyter_client/pull/314
    "ipywidgets",
    "matplotlib",
    "pyvista",
]

testing_deps = [
    "flake8",
    "pytest",
    "hypothesis",
    "affine",
    "pytest_cases",
    "subsurface",
]

extras_require = {
    "notebook": notebook_deps,
    "testing": testing_deps,
    "docs": [
        "sphinx",
        "sphinx_rtd_theme",
        "sphinxcontrib-programoutput",
        "nbsphinx >= 0.7",
        "pandoc",
        "nbconvert!=5.4",
        "nbformat",
        "sphinx-copybutton",
        "rtds-action",
        "jupytext",
        "pyvista",
    ]
    + notebook_deps,
}

setup(
    name="segysak",
    # version="0.2.3a",
    description="SEG-Y Seismic Data Inspection and Manipulation Tools using Xarray",
    long_description=readme,
    long_description_content_type="text/x-rst",
    author="SEGY-SAK Developers",
    author_email="segysak@gamil.com",
    url="https://github.com/trhallam/segysak",
    license="GPL3",
    install_requires=install_requires,
    tests_require=testing_deps,
    extras_require=extras_require,
    packages=find_packages(),
    include_package_data=True,
    # add command line scripts here
    entry_points={"console_scripts": ["segysak=segysak._cli:cli"]},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering",
    ],
    setup_requires=[
        "pytest-runner",
        "setuptools_scm",
    ],
    use_scm_version={
        "root": ".",
        "relative_to": __file__,
        "local_scheme": "no-local-version",
        "version_scheme": "post-release",
        "write_to_template": "__version__ = '{version}'",
        "write_to": "segysak/version.py",
    },
    python_requires=">=3.6",
)



#!/usr/bin/env python3

"""
Implement a simple compare command, primarily to check a round trip of
compress and decompress.
"""

import numpy as np
import os, sys, argparse
import matplotlib.pyplot as plt
from ..api import ZgyReader, ZgyWriter, SampleDataType, ProgressWithDots
from ..impl.enum import BrickStatus
from ..impl.compress import CompressStats as Stats
from ..test.utils import SDCredentials
from ..iterator import readall

def compare_stats(r1, r2):
    old = r1.statistics
    new = r2.statistics
    old_avg = old.sum / old.cnt
    old_rms = float(np.sqrt(old.ssq / old.cnt))
    new_avg = new.sum / new.cnt
    new_rms = float(np.sqrt(new.ssq / new.cnt))
    # Note, use "rms" as the reference also for avg, as old_avg may be ~0.
    delta_avg = (new_avg - old_avg) / old_rms
    delta_rms = (new_rms - old_rms) / old_rms
    # Also check the value range.
    delta_min = abs(old.min - new.min) / (old.max - old.min)
    delta_max = abs(old.max - new.max) / (old.max - old.min)
    delta_range = max(delta_min, delta_max)
    print("in: ", old)
    print("out:", new)
    print('"Difference in stats" "avg:" {0:7.4f}% "rms:" {1:7.4f}% "range:" {2:7.4f}% "count:" {3}'.format(
        100*delta_avg, 100*delta_rms, 100*delta_range, new.cnt - old.cnt))

def compare_open_files(r1, r2, w, progress, histogram, label):
    def _roundup(x, step): return ((x + step - 1) // step) * step
    filesize1 = r1._fd.xx_eof
    filesize2 = r2._fd.xx_eof
    stats = Stats("MyStats")
    bricksize = np.minimum(np.array(r1.bricksize, dtype=np.int64),
                           np.array(r2.bricksize, dtype=np.int64))
    bufsize = np.zeros(3, dtype=np.int64)
    delta_histogram_clip = 0.01 # @@ change me!
    delta_histogram = None
    delta_histogram = np.zeros(200, dtype=np.int64)
    input_rms = np.sqrt(r1.statistics.ssq / r1.statistics.cnt)
    delta_hist_range = (-input_rms * delta_histogram_clip,
                        +input_rms * delta_histogram_clip)
    for start, count, _ in readall(r1, sizeonly=True):
        bufsize = np.maximum(bufsize, count)
    data1 = np.zeros(bufsize, dtype=np.float32)
    data2 = np.zeros(bufsize, dtype=np.float32)
    for start, count, _ in readall(r1, sizeonly=True, progress=progress):
        view1 = data1[:count[0],:count[1],:count[2]]
        view2 = data2[:count[0],:count[1],:count[2]]
        r1.read(start, view1)
        r2.read(start, view2)
        if w:
            w.write(start, view2 - view1)
        ddelta = (view2 - view1)
        np.clip(ddelta, delta_hist_range[0], delta_hist_range[1], out=ddelta)
        dhist = np.histogram(ddelta, bins=len(delta_histogram), range=delta_hist_range)
        delta_histogram += dhist[0]
        # readall() can't help with the inner loop because we had to
        # read from two files, not just one.
        for ii in range(0, count[0], bricksize[0]):
            for jj in range(0, count[1], bricksize[1]):
                for kk in range(0, count[2], bricksize[2]):
                    beg = np.array((ii, jj, kk), dtype=np.int64)
                    end = np.minimum(beg + bricksize, count)
                    bpos = beg // r2.bricksize
                    bstatus, _, _, bsize = r2._accessor._getBrickFilePosition(*bpos, 0)
                    d1 = view1[beg[0]:end[0],beg[1]:end[1],beg[2]:end[2]]
                    d2 = view2[beg[0]:end[0],beg[1]:end[1],beg[2]:end[2]]
                    if bstatus in (BrickStatus.Normal, BrickStatus.Compressed):
                        stats.add_data(d1, bsize, d2, msg=str(bstatus))
    MEGA = float(1024*1024)
    print("Actual file size {0:.0f} MB / {1:.0f} MB = ratio {2:.1f} relative {3:.1f}%".format(
          filesize1 / MEGA, filesize2 / MEGA, filesize1 / filesize2, 100 * (filesize2 / filesize1)))
    compare_stats(r1, r2)
    stats.dump()
    if histogram:
        #print(stats._all_info)
        # For the histogram I want to also include the lossless bricks;
        # unlike what is done in CompressStats.dump().
        allsnr = list([x[0] for x in stats._all_info]) or [99.0]
        allsnr = np.clip(allsnr, 0, 99)
        allrat = list([100.0/e[1] for e in stats._all_info if e[1] != 0]) or [1]
        allrat = np.clip(allrat, 0.8001, 99.9999)
        allrat = 100.0 / allrat # Now in range 1% .. 125%
        fig, (ax1, ax2, ax3) = plt.subplots(3, figsize=(8, 9), dpi=100)
        fig.subplots_adjust(hspace=0.5)

        ax1.hist(allsnr, bins=100, range=(0, 100))
        #ax1.title.set_text("Signal to Noise (dB) for each brick.")
        #ax1.text(0.5, 0.9, 'Signal to Noise (dB) for each brick.',
        #         transform=ax1.transAxes, ha="center")
        ax1.set_xlabel('Measured signal to noise (dB)')
        ax1.set_ylabel('% bricks with this SNR')

        ax2.hist(allrat, bins=100, range=(1, 125))
        #ax2.title.set_text("Compression ratio for each brick.")
        #ax2.text(0.5, 0.9, 'Compression ratio for each brick.',
        #         transform=ax2.transAxes, ha="center")
        ax2.set_xlabel('Size in % of input')
        ax2.set_ylabel('% bricks with this size')

        xbins = len(delta_histogram)
        xaxis = np.arange(xbins, dtype=np.float32) # 0..nbins
        xaxis /= xbins # 0..1
        xaxis = 2*xaxis - 1 # -1..+1
        xaxis *= delta_histogram_clip
        #print("@@@ sample count", np.sum(delta_histogram), "zeros", delta_histogram[xbins//2])
        ax3.plot(xaxis, delta_histogram * (100 / np.sum(delta_histogram)), color='orange')
        #ax3.title.set_text("Histogram of errors, relative to RMS of data.")
        #ax3.text(0.5, 0.9, 'Histogram of errors, relative to RMS of data.',
        #         transform=ax3.transAxes, ha="center")
        ax3.set_xlabel('Error relative to RMS of data ({0:.0f})'.format(input_rms))
        ax3.set_ylabel('% samples with this error')

        fig.suptitle(label)
        plt.savefig(histogram)

def compare(filename1, filename2, outfilename, histogram, label):
    with ZgyReader(filename1, iocontext = SDCredentials()) as r1:
        with ZgyReader(filename2, iocontext = SDCredentials()) as r2:
            if r1.size != r2.size:
                print("Survey size mismatch:", r1.size, r2.size)
            else:
                if outfilename:
                    with ZgyWriter(outfilename,
                                   templatename=filename1,
                                   datatype=SampleDataType.float,
                                   iocontext = SDCredentials()) as w:
                        compare_open_files(r1, r2, w, progress=ProgressWithDots(), histogram=histogram, label=label)
                        w.finalize(progress=ProgressWithDots())
                else:
                    compare_open_files(r1, r2, None, progress=ProgressWithDots(), histogram=histogram, label=label)

if __name__ == "__main__":
    np.seterr(all='raise')
    parser = argparse.ArgumentParser(description='Compare two presumed equal ZGY files. Report any noise that might be caused by compression and/or quantization.', epilog=None)
    parser.add_argument('input1', help='ZGY input cube, local or sd://')
    parser.add_argument('input2', help='ZGY input cube, local or sd://')
    parser.add_argument('--output', help='Optional difference ZGY output cube, local or sd://, overwritten if it already exists.')
    parser.add_argument('--histogram', help='Optional histograms of noise and compression saved to this png file. Local files only.')
    parser.add_argument('--label', help='Title for the histogram. Defaults to the file names.')
    args = parser.parse_args()
    #print(args)
    if not args.label:
        args.label = "File 1: {0}\nFile 2: {1}".format(args.input1, args.input2)
    compare(args.input1, args.input2, args.output, args.histogram, args.label)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

"""
This file contains the public OpenZGY API.

The API is modeled roughly after the API exposed by the Python wrapper
around the existing C++ ZGY-Public API. This is probably just as good
a starting point than anything else. And it makes testing simpler for
those tests that compare the old and new behavior.

    api.ZgyMeta:

        * Has a private reference to a impl.meta.ZgyInternalMeta instance.
        * Contains a large number of properties exposing meta data,
          most of which will present information from the ZgyInternalMeta
          in a way that is simpler to use and doesn't depend on the
          file version.
        * End users will access methods from this class. Most likely
          via the derived ZgyReader and ZgyWriter classes. The end users
          might not care that there is a separate base class.

    api.ZgyMetaAndTools(ZgyMeta):

        * Add coordinate conversion routines to a ZgyMeta instance.

    api.ZgyReader(ZgyMetaAndTools): # + read, readconst, close
    api.ZgyWriter(ZgyMetaAndTools): # + write, writeconst, finalize, close

        * Has a private reference to a impl.meta.ZgyInternalBulk instance.
        * Add bulk read and write functionality, forwarding the requests
          to the internal bulk instance.
        * These classes with their own and inherited methods and properties
          comprise the public OpenZGY API.
        * Currently the ZgyWriter does not expose the bulk read() function
          but it does allow accessing all the metadata. Allowing read()
          might be added later if it appears to be useful.
          In practice this just means to let ZgyWriter inherit ZgyReader.
"""

##@package openzgy
#@brief The top level only has package members.
##@package openzgy.api
#@brief User visible apis are here.

##
# \mainpage
#
# The %OpenZGY Python API allows read/write access to files
# stored in the ZGY format. The main part of the API is here:
#
# \li ZgyReader and its ZgyMeta base class.
# \li ZgyWriter also extending ZgyMeta.
# \li ZgyUtils for anything not read or write.
# \li \ref exception.ZgyError
#  you might want to catch.
# \li ProgressWithDots example of progress reporting.
# \li \ref Example Example application.
#
# If you are reading this document from doxygen/pure/apidoc.pdf
# in the source tree then please see doxygen/README.md for an
# explanation of why the documentation produced by the build might
# be better.
#
# \if IMPL
# If you are viewing the full Doxygen documentation then this
# covers both the API and most of the implementation. So if you
# look at the list of classes and methods this might seem a bit
# daunting. All you really need to use the API should be in the
# above list. Excluding trivial structs that will be cross
# referenced as needed. So you don't need to go looking for them.
# Of course, if you want to work om %OpenZGY itself then you
# probably need everything.
#
# See also the following related pages:
#
# \li \ref physicalformat
# \li \ref implementation
# \li \ref lowres
# \li \ref migration
#
# \endif
# \page Example
# \include simplecopy.py

import numpy as np
import json
import sys
from collections import namedtuple
from enum import Enum

from .exception import *
from .impl import enum as impl_enum
from .impl.meta import ZgyInternalMeta
from .impl.bulk import ZgyInternalBulk, ScalarBuffer
from .impl.transform import generalTransform
from .impl.file import FileFactory
from .impl.stats import StatisticData
from .impl.lodalgo import DecimationType
from .impl.compress import CompressFactoryImpl
from .impl.zfp_compress import ZfpCompressPlugin
from .impl.genlod import GenLodC

##@cond IMPL

##@brief Explicit control of imported symbols.
class _internal:
    """
    This class is only used to rename some internal code.

    I want to be explicit about which classes I need from impl.*
    but at the same time I don't want to pollute the user-visible api
    namespace with names not even starting with an underscore that the
    user has no business accessing. I am not sure whether my way of
    achieving this is considered a kludge.

    What I really want is to say something like:

        from .impl.genlod import GenLodC as _internal.GenLodC

    but apparently this is not allowed.
    """
    pass
_internal.enum             = impl_enum;        del impl_enum
_internal.ZgyInternalMeta  = ZgyInternalMeta;  del ZgyInternalMeta
_internal.ZgyInternalBulk  = ZgyInternalBulk;  del ZgyInternalBulk
_internal.ScalarBuffer     = ScalarBuffer;     del ScalarBuffer
_internal.generalTransform = generalTransform; del generalTransform
_internal.FileFactory      = FileFactory;      del FileFactory
_internal.StatisticData    = StatisticData;    del StatisticData
_internal.DecimationType   = DecimationType;   del DecimationType
_internal.CompressFactoryImpl = CompressFactoryImpl; del CompressFactoryImpl
_internal.ZfpCompressPlugin   = ZfpCompressPlugin;   del ZfpCompressPlugin
_internal.GenLodC          = GenLodC;          del GenLodC

##@endcond

##@brief Sample data type used in the public API.
class SampleDataType(Enum):
    """
    Sample data type used in the public API.
    Corresponds to RawDataType used in the ZGY file format.
    """
    unknown = 1000
    int8    = 1001
    int16   = 1002
    float   = 1003

##@brief Horizontal or vertical dimension as used in the public API.
class UnitDimension(Enum):
    """
    Horizontal or vertical dimension as used in the public API.

    Horizontal dimension may be length or arc angle, although most
    applications only support length. Vertical dimension may be time
    or length. Vertical length is of course the same as depth.
    Arguably there should have been separate enums for horizontal and
    vertical dimension since the allowed values differ.
    """
    unknown  = 2000
    time     = 2001
    length   = 2002
    arcangle = 2003

##@brief Base class shared betewwn ZgyReader and ZgyWriter.
class ZgyMeta:
    """
    Base class shared betewwn ZgyReader and ZgyWriter.
    """
    def __init__(self, meta):
        """Create an instance, providing the ZgyInternalMeta to use."""
        assert meta is not None
        assert isinstance(meta, _internal.ZgyInternalMeta)
        self._meta = meta

    @property
    def size(self): # (iii)
        """
        Number of inlines, crosslines, and samples in that order.
        """
        return self._meta._ih._size

    @property
    def datatype(self): # s -> Enum
        """
        Sample data type.
        The ZGY-Public API uses enums: "int8", "int16", "float".
        In some cases these are also passed as strings.
        The old Python wrapper for ZGY-Public is stringly typed.
        Instead of returning a real enum it returns the name.
        """
        return _map_DataTypeToSampleDataType(self._meta._ih._datatype)

    @property
    def datarange(self): # (ff), a.k.a. DataMinMax
        """
        For integral data this is the lowest and highest sample value
        than can be represented in storage. The lowest storage value
        (e.g. -128 for SignedInt8 data) will be returned as DataMinMax[0]
        when read as float. Similarly the highest storage value e.g. +127
        will be returned as DataMinMax[1]. When integer data is read as
        the "native" integral type then no automatic scaling is applied.
        Note that in this case the actual range of the samples on file might
        be smaller (for int8, not all of the range -128..+127 might be used)
        but it cannot be larger.

        For floating point data these numbers are supposed to be the actual
        value range of the samples on file. It is a good idea to enforce
        this here, as the values stored by older writers cannot be trusted.
        Note: Also enforced on write in impl.meta.InfoHeaderV2.calculate_write.
        TODO-Worry: In some float32 datasets the bulk data might have
        ridiculously large spikes wich will be included in the statistical
        range but not in the codingrange. So, codingrange is actually the
        one that is correct. Also, can we have a situation where stats
        are not filled in while the codingrange is set? I am not sure
        this is currently handled.
        """
        if self._meta._ih._datatype == _internal.enum.RawDataType.Float32:
            return (self._meta._ih._smin, self._meta._ih._smax)
        else:
            return (self._meta._ih._safe_codingrange[0], self._meta._ih._safe_codingrange[1])

    @property
    def raw_datarange(self): # (ff), a.k.a. DataMinMax
        """
        As datarange, but the actual values read from the file before
        they might have been changed to try to fix a bad file.
        Only use this property if you want to handle such files
        differently than the library does.
        """
        if self._meta._ih._datatype == _internal.enum.RawDataType.Float32:
            return (self._meta._ih._smin, self._meta._ih._smax)
        else:
            return (self._meta._ih._file_codingrange[0], self._meta._ih._file_codingrange[1])

    @property
    def zunitdim(self):
        """
        Dimension in the vertical direction. "time" or "length".
        "time" might on file be "SeismicTWT" or "SeismicOWT".
        The old Python wrapper for ZGY-Public is stringly typed.
        Instead of returning a real enum it returns the name.
        """
        return _map_VerticalDimensionToUnitDimension(self._meta._ih._vdim)

    @property
    def hunitdim(self):
        """
        Dimension in the horizontal direction. Should always be "length".
        The original specification called for supporting "arcangle" as well,
        i.e. coordinates in degrees instead of a projection. But most
        application code that use ZGY will not support this.
        The old Python wrapper for ZGY-Public is stringly typed.
        Instead of returning a real enum it returns the name.
        """
        return _map_HorizontalDimensionToUnitDimension(self._meta._ih._hdim)

    @property
    def zunitname(self):
        """
        Unit in the horizontal direction. E.g. "ms", "m", or "ft".
        Note that Petrel might ignore this settings and instead
        prompt the user to state what the unit should be.
        """
        return self._meta._ih._vunitname

    @property
    def hunitname(self):
        """
        Unit in the horizontal direction. E.g. "m" or "ft".
        """
        return self._meta._ih._hunitname

    @property
    def zunitfactor(self):
        """
        Factor to multiply stored vertical values with to get SI units.
        E.g. 0.001 for ms, 1.0 for m or 0.3048 for ft.
        """
        return self._meta._ih._vunitfactor

    @property
    def hunitfactor(self):
        """
        Factor to multiply stored horizontal values with to get SI units.
        """
        return self._meta._ih._hunitfactor

    @property
    def zstart(self):
        """
        Distance from surface/MSL to first sample, given in the vertical unit.
        """
        return self._meta._ih._orig[2]

    @property
    def zinc(self):
        """
        Sample interval, given in the vertical unit.
        """
        return self._meta._ih._inc[2]

    @property
    def annotstart(self):
        """
        First inline and crossline numbers.
        """
        return self._meta._ih._orig[0:2]

    @property
    def annotinc(self):
        """
        Inline and crossline number increments between adjacent
        sections of the cube.
        """
        return self._meta._ih._inc[0:2]

    @property
    def corners(self):
        """
        World XY coordinates of each of the 4 corners.
        The same coordinates in ordinal numbers are
        ((0, 0), (Size[0]-1, 0), (0, Size[1]-1), (Size[0]-1, Size[0]-1))
        """
        return self._meta._ih._ocp_world

    @property
    def indexcorners(self):
        """
        Redundant with Size.
        Ordinal coordinates of each of the 4 corners, ordered as "corners".
        """
        return self._meta._ih._ocp_index

    @property
    def annotcorners(self):
        """
        Redundant with Start, Inc, Size.
        Annotation coordinates of each of the 4 corners, ordered as HCorners.
        """
        return self._meta._ih._ocp_annot

    @property
    def bricksize(self):
        """
        Size of a brick. Should always be (64, 64, 64).
        """
        return self._meta._ih._bricksize

    @property
    def brickcount(self):
        """
        Number of bricks (including empties) ordered by [lod][dimension].
        """
        return self._meta._ih._lodsizes

    @property
    def nlods(self):
        """
        Number of level-of-detail layers, including lod 0 a.k.a. full resolution.
        """
        return self._meta._ih._nlods

    @staticmethod
    def _formatUUID(uuid):
        """
        Convert a little-endian binary UUID to a big-endian string version.
        See the C++ version for details.

        First part byteswaps as an uint32_t.
        Second and third part byteswaps as two uint16_t.
        Remaining two parts are not byteswapped.
        Hyphens added between parts.
        """
        return ("{3:02x}{2:02x}{1:02x}{0:02x}-" +
                "{5:02x}{4:02x}-{7:02x}{6:02x}-" +
                "{8:02x}{9:02x}-" +
                "{10:02x}{11:02x}{12:02x}{13:02x}{14:02x}{15:02x}").format(*uuid)

    #@property
    #def dataid(self):
    #    """
    #    GUID set on file creation.
    #    """
    #    return self._formatUUID(self._meta._ih._dataid)

    @property
    def verid(self):
        """
        GUID set each time the file is changed.
        """
        return self._formatUUID(self._meta._ih._verid)

    #@property
    #def previd(self):
    #    """
    #    GUID before last change.
    #    """
    #    return self._formatUUID(self._meta._ih._previd)

    @property
    def meta(self):
        """
        A dictionary of all the meta information, which can
        later be passed as **kwargs to the ZgyWriter constructor.
        and "indexcorners", "annotcorners", "brickcount", "nlods"
        are all derived properties that will never be settable.
        "numthreads" is a property of the implementation, not the file.
        """
        return {
            "size":        self.size,
            "bricksize":   self.bricksize,
            "datatype":    self.datatype,
            "datarange":   self.datarange,
            "zunitdim":    self.zunitdim,
            "hunitdim":    self.hunitdim,
            "zunitname":   self.zunitname,
            "hunitname":   self.hunitname,
            "zunitfactor": self.zunitfactor,
            "hunitfactor": self.hunitfactor,
            "zstart":      self.zstart,
            "zinc":        self.zinc,
            "annotstart":  self.annotstart,
            "annotinc":    self.annotinc,
            "corners":     self.corners,
        }

    @property
    def numthreads(self):
        """
        How many threads to use when reading. Currently ignored.
        """
        return 1
    @numthreads.setter
    def numthreads(self, x):
        print("Warning: numthreads is ignored.")

    def dump(self, file=None):
        file = file or sys.stdout
        print("{", file=file)
        for e in sorted(self.meta.items()):
            value = '"'+e[1]+'"' if isinstance(e[1], str) else str(e[1])
            print('  "{0}": {1},'.format(e[0], value), file=file)
        print("}", file=file)

    ### New in OpenZGY ###
    _statisticsType = namedtuple("Statistics", "cnt sum ssq min max")
    @property
    def statistics(self):
        """
        Return the statistics stored in the file header as a named tuple.
        NOTE, I might want to change this to another type if there is a
        need to implement the same method in the ZGY-Public wrapper,
        as it might be trickier to define a namedtuple there.
        """
        return self._statisticsType(self._meta._ih._scnt,
                                    self._meta._ih._ssum,
                                    self._meta._ih._sssq,
                                    self._meta._ih._smin,
                                    self._meta._ih._smax)

    _histogramType = namedtuple("Histogram", "cnt min max bin")
    @property
    def histogram(self):
        """
        Return the statistics stored in the file header as a named tuple.
        NOTE, I might want to change this to another type if there is a
        need to implement the same method in the ZGY-Public wrapper,
        as it might be trickier to define a namedtuple there.
        """
        if not self._meta._hh: return None
        return self._histogramType(self._meta._hh._cnt,
                                   self._meta._hh._min,
                                   self._meta._hh._max,
                                   self._meta._hh._bin)

##@brief Base class shared betewwn ZgyReader and ZgyWriter.
class ZgyMetaAndTools(ZgyMeta):
    """
    Base class shared betewwn ZgyReader and ZgyWriter.
    Adds coordinate conversion tools.
    """

    @staticmethod
    def transform(A, B, data):
        """
        Linear transformation of an array of double-precision coordinates.
        The coordinate systems to convert between are defined by
        three arbitrary points in the source system and the target.
        Arguments: ((ax0,ay0), (ax1,ay1), (ax2,ay2)),
                   ((bx0,by0), (bx1,by1), (bx2,by2)),
                   data
        where data is a 2d array of size (length, 2)
        """
        # performance note: In Python it can be far more efficient to
        # build and cache 6 transformation matrices between index/annot/world
        # and use those for the 6 transforms. But if we only transform
        # a few values at a time anyway, or if we are planning to convert
        # the accessor back to C++ fairly soon, this is a non-issue.
        _internal.generalTransform(
            A[0][0], A[0][1], A[1][0], A[1][1], A[2][0], A[2][1],
            B[0][0], B[0][1], B[1][0], B[1][1], B[2][0], B[2][1],
            data)

    @staticmethod
    def transform1(A, B, point):
        data = [[point[0], point[1]]]
        ZgyMetaAndTools.transform(A, B, data)
        return tuple(data[0])

    def annotToIndex(self, point):
        """Convert inline, crossline to ordinal"""
        return self.transform1(self.annotcorners, self.indexcorners, point)

    def annotToWorld(self, point):
        """Convert inline, crossline to world X,Y"""
        return self.transform1(self.annotcorners, self.corners, point)

    def indexToAnnot(self, point):
        """Convert ordinal to inline, crossline"""
        return self.transform1(self.indexcorners, self.annotcorners, point)

    def indexToWorld(self, point):
        """Convert ordinal to world X,Y"""
        return self.transform1(self.indexcorners, self.corners, point)

    def worldToAnnot(self, point):
        """Convert world X,Y to inline, crossline"""
        return self.transform1(self.corners, self.annotcorners, point)

    def worldToIndex(self, point):
        """Convert world X,Y to ordinal"""
        return self.transform1(self.corners, self.indexcorners, point)

##@brief Main entry point for reading ZGY files.
class ZgyReader(ZgyMetaAndTools):
    """
    Main entry point for reading ZGY files.

    Obtain a concrete instance by calling the constructor.
    You can then use the instance to read both meta data and bulk data.
    It is recommended to explicitly close the file when done with it.
    """
    def __init__(self, filename, *, _update=False, iocontext=None):
        # No "with" statement for the FileFactory, so we must remember
        # to close it ourselves in our own __exit__.
        self._fd = _internal.FileFactory(filename, ("r+b" if _update else "rb"), iocontext)
        self._meta = _internal.ZgyInternalMeta(self._fd)
        # self._meta._ih and friends will all be allocated.
        # Prove that all the tests for "._ih is not None" are redundant.
        self._meta._assert_all_headers_allocated()
        # At the implementation level the bulk- and meta access are separate,
        # and the bulk accessor needs some of the meta information to work.
        self._accessor = _internal.ZgyInternalBulk(self._fd, self._meta)
        # This causes an assignment to the parent's self._meta
        # which in Python is a no-op but in C++ the parent might
        # have its own _meta that we shadow here. Or not.
        super().__init__(self._meta)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        # Note that if the block was exited due to an exception, and if we
        # also get an exception from close, then it is the second one
        # that gets caught in a try/catch block placed outside the "while".
        # Callers will in this case often want to report just the first
        # exception since the close() probably failed as a cause of it.
        # Caller needs to do e.g. "ex2 = ex.__cause__ or ex.__context__".
        # It is possible to instead suppress any errors from close if
        # another exception is already pending. Simplifying the caller.
        # But then the close() would not be available at all. Bad idea.
        self.close()

    ##@brief Read bulk data into a caller specified buffer.
    def read(self, start, data, *, lod = 0, verbose = None):
        """
        Read an arbitraty region of bulk data into a caller specified buffer.

        The buffer's type must be float, short, or char. Any file may
        be read as float. If the buffer is of type short or char then
        the file must be of precisely that type.

        Arguments: (i0,j0,k0), buffer, lod=0
        """
        file_dtype = _internal.enum._map_DataTypeToNumpyType(self._meta._ih._datatype)
        if (not isinstance(data, np.ndarray) or
            data.dtype not in (np.float32, file_dtype) or
            len(data.shape) != 3 or
            not data.flags.writeable
        ):
            raise ZgyUserError("Expected a writeable 3d numpy array of np.{0} or np.{1}".format(
                np.dtype(np.float32).name, np.dtype(file_dtype).name))
        if not self._accessor: raise ZgyUserError("ZGY file is not open for read.")
        self._accessor.readToExistingBuffer(data, start, lod=lod,
                                            as_float=(data.dtype==np.float32),
                                            verbose=verbose)

    ##@brief Get hint about all constant region.
    ##@image html readconst-fig1.png
    ##@image latex readconst-fig1.png
    def readconst(self, start, size, *, lod = 0, as_float = True, verbose = None):
        """
        Get hint about all constant region.

        Check to see if the specified region is known to have all samples
        set to the same value. Returns that value, or None if it isn't.

        The function only makes inexpensive checks so it might return
        None even if the region was in fact constant. It will not make
        the opposite mistake. This method is only intended as a hint
        to improve performance.

        For int8 and int16 files the caller may specify whether to scale
        the values or not.
        """
        if not self._accessor: raise ZgyUserError("ZGY file not open for read")
        return self._accessor.readConstantValue(start, size, lod=lod,
                                                as_float=as_float,
                                                verbose=verbose)

    def close(self):
        if self._fd:
            self._fd.xx_close()
            self._fd = None
        if self._accessor:
            self._accessor = None # No other cleanup needed.
        # Metadata remains accessible. Not sure whether this is a good idea.


##@brief Main API for creating ZGY files.
class ZgyWriter(ZgyMetaAndTools):
    """
    Main API for creating ZGY files.

    Obtain a concrete instance by calling the constructor.
    All meta data is specified in the call to open(), so meta data
    will appear to be read only. You can use the instance to write
    bulk data. The file becomes read only once the instance is closed.

    It is recommended to call finalize() and close() after all bulk
    has been written. But if you forget then this will be done when
    the writer goes out of scope, provided of course that you used a
    "with" block.
    """
    def __init__(self, filename, *,
                 iocontext=None, compressor=None, lodcompressor=None, **kwargs):
        """
        Create a new ZGY file.

        Optionally pass templatename = otherfile to create a new file
        similar to otherfile. Explicit keyword argumants override
        information from otherfile.

        Optionally pass templatename = filename to erase all data blocks
        from filename but keep the metadata. New data blocks can then
        be written to the file. Petrel/BASE might need this feature,
        due to the way it writes new files. They tend to get opened
        several times to add meta information. Caveat: Behind the
        scenes the file is simply deleted and re-created. This is
        slightly less efficient than opening the file for read/write.

        templatename: string
            Optionally create a file similar to this one.
            TODO-Low: In the future might also accept a ZgyReader instance.
            This is convenient if a file is being copied, so as to not
            needing to open it twice.

        filename: string
            The local or cloud file to create.

        size: (int, int, int)
            Number of inlines, crosslines, samples.

        bricksize: (int, int, int)
            Size of a single brick. Defaults to (64, 64, 64).
            Please use the default unless you really know what
            you are doing. In any case, each size needs to be
            a power of 2.

        datatype: SampleDataType
            Specify int8, int16, or float.

        datarange = (float, float)
            Used only if datatype is integral, to convert from storage to
            actual sample values. The lowest possible storage value, i.e.
            -128 or -32768, maps to datarange[0] and the highest possible
            storage value maps to datarange[1].

        zunitdim:    UnitDimension. time, length, or unknown.
        zunitname:   string
        zunitfactor: float
            Describe how to convert between storage units and SI units
            in the vertical direction. Petrel ignores these settings and
            prompts the user.

        hunitdim:    UnitDimension. length, arcangle, or unknown.
        hunitname:   string
        hunitfactor: float
            Describe how to convert between storage units and SI units
            in the horizontal direction. Most applications cannot handle
            arcangle. Petrel ignores these settings and prompts the user.

        zstart: float
            The time or depth corresponding to the shallowest sample.

        zinc: float
            The vertical time or depth distance between neighboring samples.

        annotstart: (float, float)
            The inline and crossline numbers corresponding to the ordinal
            position (0, 0) i.e. the first sample on the file.

        annotinc: (float, float)
            The inline / crossline step between neighboring samples.
            The samples at ordinal (1, 1) will have annotation
            annotstart + annotinc.

        corners: (float, float)[4]
            World coordinates of each corner, order as
                First inline / first crossline,
                last inline / first crossline,
                first inline / last crossline,
                last inline / last crossline.

        compressor, lodcompressor: callable
            If set, attempt to compress each block with this callable.
            Typically this should be a lambda or a class, because it
            needs to capture the compression parameters.
            Example:
                compressor = ZgyCompressFactory("ZFP", snr = 30)
            If different compression parameters are desired for
            full- and low resolution bricks then lodcompressor can
            be provided as well. It defaults to compressor. Using
            two different instances, even if the parameters match,
            may also cause statistics to be reported separately
            for fullres and lowres.
            TODO-Low: Future: passing zfp_compressor = snr is equivalent
            to compressor = ZgyCompressFactory("ZFP", snr = snr).
            Unlike the compressor keyword this also works in the wrapper.
        """
        # The following arguments are not passed on to _create:
        #   -  templatename is handled locally, using the template for defaults.
        #   -  iocontext is only made available to the FileADT layer
        #   -  compressor is explicitly passed as an argument to those
        #      functions (write, writeconst, finalize) that need it.

        if "templatename" in kwargs:
            with ZgyReader(kwargs["templatename"]) as t:
                for k, v in t.meta.items():
                    if not k in kwargs:
                        kwargs[k] = v
            del kwargs["templatename"]

        # Compressing a file as integer is not useful, it just adds more noise
        # at least as long as we are talking about seismic. Even if we allowed
        # int8 or int16 here the code in impl_zfp_compress will currently
        # use the ZFP float interface.
        #
        # Flagging as int8 or int16 would save memory at the cost of adding
        # even more noise. Converting the decompressed data to integral before
        # returning it as that type. But if the applicaton wants this then
        # it can easily convert the result itself.
        #
        # There are other subtle issues with int8 / int16. Even with enabled
        # compression, individual bricks are allowed to be stored uncompressed
        # if neither lossy nor lossless compression works as desired. The
        # declared value type then controls how these blocks are stored.
        # Storing those (presumably very few) blocks as integral means
        # more that can go wrong and more unit  tests. And keep in mind
        # that float->int8 and float->int16 are also a form of compression
        # (4x and 2x respectively) but is a lot noisier than ZFP at the same
        # compression factor.
        #
        # We may need to revisit this if switching to another compression
        # algorithm where integral compression works better.
        if compressor and kwargs.get("datatype", SampleDataType.float) != SampleDataType.float:
            raise ZgyUserError("Compressed files need to be stored as float.")

        # After this, self._meta._ih and friends exists but will be None.
        self._meta = _internal.ZgyInternalMeta(None)
        # This causes an assignment to the parent's self._meta
        # which in Python is a no-op but in C++ the parent might
        # have its own _meta that we shadow here. Or not.
        super().__init__(self._meta)
        self._create(filename, compressed = bool(compressor or lodcompressor), **kwargs)
        # Now self._meta._ih and friends will all be allocated.
        # Prove that all the tests for "._ih is not None" are redundant.
        self._meta._assert_all_headers_allocated()
        # The file creation was deferred until after the consistency checks.
        # No "with" statement for the FileFactory, so we must remember
        # to close it ourself in our own __exit__.
        self._fd = _internal.FileFactory(filename, "w+b", iocontext)
        self._meta._flush_meta(self._fd)
        # The accessor needs to know whether we will do compression or not,
        # because this determines whether bricks will be written aligned
        # and possibly whether updates are allowed. The actual
        # compression algorithm is passed on each write etc.
        # TODO-Low consider storing the compressor as context of the accessor
        # instead. Less precise control though. We might want a different
        # snr on the low resolution bricks.
        self._accessor = _internal.ZgyInternalBulk(
            self._fd, self._meta,
            compressed = bool(compressor or lodcompressor))
        self._dirty = False # If True need LOD, stats, histogram.
        self._compressor = compressor or lodcompressor
        self._lodcompressor = lodcompressor or compressor

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        # Note that if the block was exited due to an exception, and if we
        # also get an exception from close, then it is the second one
        # that gets caught in a try/catch block placed outside the "while".
        # Callers will in this case often want to report just the first
        # exception since the close() probably failed as a cause of it.
        # Caller needs to do e.g. "ex2 = ex.__cause__ or ex.__context__".
        # This is mitigated by the close() method skipping work and/or
        # suppressing exceptions if the file has already been flagged bad.
        self.close()

    def _create(self, filename, *, size = None, compressed = False,
                bricksize = None,
                datatype = SampleDataType.float, datarange = None,
                zunitdim = UnitDimension.unknown,
                hunitdim = UnitDimension.unknown,
                zunitname = None, hunitname = None,
                zunitfactor = 0.0, hunitfactor = 0.0, zstart = 0.0, zinc = 0.0,
                annotstart = (0, 0), annotinc = (0, 0),
                corners = ((0,0),(0,0),(0,0),(0,0))):
        """
        Called from __init__. Do not call directly.
        The user should use a "using" statement when creating the reader.
        Datatype can be "int8", "int16", "float".
        Dimension can be "time", "length", or "arcangle".
        """
        self._meta._init_from_scratch(filename = filename,
                                size = size,
                                compressed = compressed, # If True this is V4.
                                bricksize = bricksize,
                                datatype = _map_SampleDataTypeToDataType(datatype),
                                datarange = datarange,
                                zunitdim = _map_UnitDimensionToVerticalDimension(zunitdim),
                                hunitdim = _map_UnitDimensionToHorizontalDimension(hunitdim),
                                zunitname = zunitname,
                                hunitname = hunitname,
                                zunitfactor = zunitfactor,
                                hunitfactor = hunitfactor,
                                zstart = zstart,
                                zinc = zinc,
                                annotstart = annotstart,
                                annotinc = annotinc,
                                corners = corners)

    ##@brief Write an arbitrary region.
    def write(self, start, data, *, verbose = None):
        """
        Write bulk data. Type must be np.float32, np.int16, or np.int8.
        np.float32 may be written to any file and will be converted
        if needed before storing. Writing an integral type implies that
        values are to be written without conversion. In that case the
        type of the buffer must match exactly the file's storage type.
        You cannot write int8 data to an int16 file or vice versa.

        Arguments:
            start: tuple(i0,j0,k0) where to start writing.
            data: np.ndarray of np.int8, np.int16, or np.float32
        """
        file_dtype = _internal.enum._map_DataTypeToNumpyType(self._meta._ih._datatype)
        if (not isinstance(data, np.ndarray) or
            data.dtype not in (np.float32, file_dtype) or
            len(data.shape) != 3
        ):
            raise ZgyUserError("Expected a 3d numpy array of np.{0} or np.{1}".format(
                np.dtype(np.float32).name, np.dtype(file_dtype).name))
        if not self._accessor: raise ZgyUserError("ZGY file is not open.")
        if self._accessor._is_bad or self._meta._is_bad:
            raise ZgyCorruptedFile("Cannot continue due to previous errors.")
        self._accessor._writeRegion(data, start, lod=0,
                                    compressor=self._compressor,
                                    is_storage=(data.dtype != np.float32),
                                    verbose=verbose)
        self._dirty = True

    ##@brief Write a single value to a region of the file.
    def writeconst(self, start, value, size, is_storage, *, verbose = None):
        """
        Write a single value to a region of the file.

        This is equivalent to creating a constant-value array with np.full()
        and write that. But this version might be considerably faster.

        If is_storage is false and the input value cannot be converted to
        storage values due to being outside range after conversion then
        the normal rules (use closest valid value) apply. If
        is_storage is True then an error is raised if the supplied value
        cannot be represented.

        Arguments:
            start:     tuple(i0,j0,k0) where to start writing.
            size:      tuple(ni,nj,nk) size of region to write.
            value:     Scalar to be written.
            is_storge: False if the value shall be converted to storage
                       True if it is already storage and should be written
                       unchanged. Ignored if the storage type is float.
        """
        if self._meta._ih._datatype == _internal.enum.RawDataType.Float32:
            dtype = np.float32 # "Convert" to user now. Actually a no-op.
            is_storage = False
        elif not is_storage:
            dtype = np.float32 # Force conversion.
        elif self._meta._ih._datatype == _internal.enum.RawDataType.SignedInt16:
            dtype = np.int16
        elif self._meta._ih._datatype == _internal.enum.RawDataType.SignedInt8:
            dtype = np.int8
        else:
            raise ZgyFormatError("Unrecognized datatype on file")
        if np.issubdtype(dtype, np.integer) and not np.isfinite(value):
            raise ZgyUserError("Cannot store {0} in a {1}".format(value, np.dtype(dtype)))
        self._accessor._writeRegion(_internal.ScalarBuffer(size, value, dtype),
                                    start, lod=0,
                                    compressor=self._compressor,
                                    is_storage=is_storage,
                                    verbose=verbose)
        self._dirty = True

    def finalize(self, *, decimation=None, progress=None, force=False, verbose=None):
        """
        Generate low resolution data, statistics, and histogram.
        This will be called automatically from close(), but in
        that case it is not possible to request a progress callback.

        If the processing raises an exception the data is still marked
        as clean. Called can force a retry by passing force=True.

        Arguments:
            decimation: Optionally override the decimation algorithms by
                        passing an array of impl.lodalgo.DecimationType
                        with one entry for each level of detail. If the
                        array is too short then the last entry is used for
                        subsequent levels.
                        TODO-Low: The expected enum type is technically internal
                        and ought to have been mapped to an enum api.XXX.
            progress:   Function(done, total) called to report progress.
                        If it returns False the computation is aborted.
                        Will be called at least one, even if there is no work.
            force:      If true, generate the low resolution data even if
                        it appears to not be needed. Use with caution.
                        Especially if writing to the cloud, where data
                        should only be written once.
            verbose:    optional function to print diagnostic information.
        """
        if self._dirty or force:
            self._dirty = False
            stats, histo = _internal.GenLodC(
                accessor   = self._accessor,
                compressor = self._lodcompressor,
                decimation = decimation,
                progress   = progress,
                verbose    = verbose)()
            # TODO-Low: Refactor:
            # violating encapsulation rather more than usual.
            # Note that _accessor._metadata is private; it is a copy
            # or actually a reference letting ZgyInternalBulk use
            # some parts of the metadata.
            (a, b) = self._accessor._scaleDataFactorsStorageToFloat()
            stats.scale(a, b)
            histo.scale(a, b)
            histo.resize(256)
            self._meta._ih._scnt = stats._cnt
            self._meta._ih._ssum = stats._sum
            self._meta._ih._sssq = stats._ssq
            self._meta._ih._smin = stats._min
            self._meta._ih._smax = stats._max
            self._meta._hh._min = histo.vv_range[0]
            self._meta._hh._max = histo.vv_range[1]
            self._meta._hh._bin = histo.bins
            self._meta._hh._cnt = np.sum(histo.bins)
        else:
            if progress:
                progress(0, 0)
        # For debugging and measurements only.
        if False:
            if self._compressor:
                self._compressor.dump(msg="Compress")
            if self._lodcompressor and not self._lodcompressor is self._compressor:
                self._lodcompressor.dump(msg="LOD_data")

        # TODO-Low: Refactor: the accessor should logically have a close(),
        # shouldn't it? And maybe I shouldn't set self._fd here.
        # Allowing the accessor to manage it.

    def close(self):
        """
        Close the currently open file.
        Failure to close the file will corrupt it.
        """
        if self._fd:
            # Both the meta accessor and the bulk accessor has an _is_bad
            # flag that is set if we got an error while writing to the file.
            # If set then don't bother with statistics, histogram, and lowres.
            # The file will probably just be discarded amyway.
            if not (self._meta._is_bad or self._accessor._is_bad):
                self.finalize()

            # Flushing metadata is a bit more ambiguous. Parts of the file
            # might still be salvageable, so especially when testing we might
            # still want to flush a file marked as bad. But, ignore any
            # secondary errors as the user already knows something is wrong.
            if not (self._meta._is_bad or self._accessor._is_bad):
                self._meta._flush_meta(self._fd)
            else:
                try:
                    self._meta._is_bad = False
                    self._accessor._is_bad = False
                    self._meta._flush_meta(self._fd)
                except Exception:
                    pass
                finally:
                    self._meta._is_bad = True
                    self._accessor._is_bad = True

            # TODO-Low it would be more consistent if the final write was
            # of the updated headers, in case the app crashes before
            # xx_close. This is true for local file access. But for
            # cloud access the last bulk data segment might be written
            # on xx_close(). Difficult to change without complicating
            # the internal FileADT api.

            # Closing the local or cloud handle is always needed as there
            # might be resources that need to be cleaned up.
            self._fd.xx_close()
            self._fd = None

            # The client code is strongly advised to delete the file if it was
            # opened for create. OpenZGY might have deleted the file itself but
            # this is probably too harsh. A file opened for update (which is not
            # actually supported yet) might still be usable. So in the future we
            # may need an additional flag telling whether the writer has ever
            # had any successful writes to the file. If not then the file is
            # still good. Note that the suggestions below need more work.
            # TODO-Low: Later: this is in the bells & whistles category.
            #if not self._precious_set_on_open:
            #    self._fd.xx_delete_on_close(); self._fd.xx_close()
            #    self._fd.xx_close_of_needed_and_delete()
            #    ZgyUtils(saved_iocontext).delete(self._filename)

    ##@cond IMPL
    @property
    def errorflag(self):
        """
        If true this means that an error happened during a critical operation
        such as a write. This means the file being written is probably bad.
        It is recommended to close and delete the file. In the future
        OpenZGY might even do that automatically.
        """
        return self._meta._is_bad or self._accessor._is_bad

    @errorflag.setter
    def errorflag(self, value):
        """
        Set or reset the flag indicating that the file got corrupted.
        """
        self._meta._is_bad = value
        self._accessor._is_bad = value
    ##@endcond

##@brief Simple progress bar.
class ProgressWithDots:
    """
    Progress bar that writes dots (51 by default) to standard output.
    This can be user as-is for simple command line apps, or you can use
    the source code as an example on how to write your own.

    The default of 51 dots will print one dot at startup and then one
    additional dot for each 2% work done.

    If you are using this to write to the cloud a file that is smaller
    than ~10 GB then the progress bar will probably move in larger
    jumps. Because writing to a cloud back-end uses very large buffers.
    Most cloud back-ends cannot report progress inside a "write block".

    When passing a progress reporter to a function, make sure you do not
    pass the class itself. You need to create an instance of it.
    """
    def __init__(self, length=51, outfile=sys.stderr):
        self._dots_printed = 0
        self._length = length
        self._outfile = outfile
    def __call__(self, done, total):
        #print("Progress: {0}/{1}".format(done, total))
        if self._dots_printed == 0:
            print("[" + (" " * self._length) + "]\r[", end='', flush=True)
        needed = 1 if not total else 1 + ((done * (self._length-1)) // total)
        if needed > self._dots_printed:
            print("." * (needed - self._dots_printed),
                  file=self._outfile, flush=True, end='')
            self._dots_printed = needed
        if done == total:
            print("", file=self._outfile)
        return True

##@brief Operations other than read and write.
class ZgyUtils:
    """
    Operations other than read and write.

    Any operations that don't fit into ZgyReader or ZgyWriter go here.
    Such as deleting a file. Or any other operation that does not need
    the file to be open first.
    """
    ##@brief Create a new concrete instance of ZgyUtils.
    def __init__(self, iocontext=None):
        """
        Create a new concrete instance of ZgyUtils.

        The reason you need to supply a file name or a file name prefix is that
        you need to provide enough information to identify the back-end that
        this instance will be bound to. So if you have registered a back-end
        named "xx", both "xx://some/bogus/file.zgy" and just "xx://" will
        produce an instance that works for your XX backend,

        For performance reasons you should consider caching one ZgyUtils
        instance for each back end you will be using. Instead of just creating
        a new one each time you want to invoke a method. Just remember that
        most operations need an instance created with the same prefix.
        """
        self._iocontext = iocontext

    ##@brief Delete a file. Works both for local and cloud files.
    def delete(self, filename):
        """
        Delete a file. Works both for local and cloud files.
        Note that the instance must be of the correct (local or cloud) type.
        """
        with _internal.FileFactory(filename, "d", self._iocontext) as f:
            f.xx_close()

def ZgyCompressFactory(name, *args, **kwargs):
    """
    Look up a compression algorithm by name and instanciate a compressor,
    passing the required compression parameters. Using this approach
    reduces the coupling between client code and the compressor.
    """
    return _internal.CompressFactoryImpl.factory(name, *args, **kwargs)

def ZgyKnownCompressors():
    """
    Return the names of all compressors known to the system.
    This is primarily for logging, but might in principle be used
    in a GUI to present a list of compressors to choose from.
    The problem with that is how to handle the argument list.
    """
    return _internal.CompressFactoryImpl.knownCompressors()

def ZgyKnownDecompressors():
    """
    Return the names of all compressors known to the system.
    This is primarily for logging.
    """
    return _internal.CompressFactoryImpl.knownDecompressors()

#############################################################################
###   Define enums used in the public API. These are separate from the    ###
###   enums used inside ZGY files, to improve isolation.                  ###
#############################################################################

def _map_DataTypeToSampleDataType(e):
    return _mapEnum(e, {
        _internal.enum.RawDataType.SignedInt8:  SampleDataType.int8,
        _internal.enum.RawDataType.SignedInt16: SampleDataType.int16,
        _internal.enum.RawDataType.Float32:     SampleDataType.float,
        None:                       SampleDataType.unknown,
    })

def _map_SampleDataTypeToDataType(e):
    return _mapEnum(e, {
        SampleDataType.int8:  _internal.enum.RawDataType.SignedInt8,
        SampleDataType.int16: _internal.enum.RawDataType.SignedInt16,
        SampleDataType.float: _internal.enum.RawDataType.Float32,
    })

def _map_HorizontalDimensionToUnitDimension(e):
    return _mapEnum(e, {
        _internal.enum.RawHorizontalDimension.Length:   UnitDimension.length,
        _internal.enum.RawHorizontalDimension.ArcAngle: UnitDimension.arcangle,
        None:                               UnitDimension.unknown,
    })

def _map_VerticalDimensionToUnitDimension(e):
    return _mapEnum(e, {
        _internal.enum.RawVerticalDimension.Depth:      UnitDimension.length,
        _internal.enum.RawVerticalDimension.SeismicTWT: UnitDimension.time,
        _internal.enum.RawVerticalDimension.SeismicOWT: UnitDimension.time,
        None:                               UnitDimension.unknown,
    })

def _map_UnitDimensionToHorizontalDimension(e):
    return _mapEnum(e, {
        UnitDimension.length:   _internal.enum.RawHorizontalDimension.Length,
        UnitDimension.arcangle: _internal.enum.RawHorizontalDimension.ArcAngle,
        UnitDimension.unknown: _internal.enum.RawHorizontalDimension.Unknown,
    })

def _map_UnitDimensionToVerticalDimension(e):
    return _mapEnum(e, {
        UnitDimension.time:   _internal.enum.RawVerticalDimension.SeismicTWT,
        UnitDimension.length: _internal.enum.RawVerticalDimension.Depth,
        UnitDimension.unknown: _internal.enum.RawVerticalDimension.Unknown,
    })

def _mapEnum(e, lookup):
    """
    Internal method to map between impl.enum tags used in the file format
    and those used in the API, to better isolate the API from changes.

    An unrecognized tag when mapping from file format to api is
    usually treated as a warning. The application might be able to
    handle it. In the lookup table, if there is an entry with key None
    then this is considered to be the default value. If there is no
    such entry then we are expected to raise an exception on error.

    When mapping in the other direction the value came
    from user code so raising an exception is usually warranted.
    """
    if e in lookup: return lookup[e]
    if None in lookup: return lookup[None]
    valid = ", ".join([str(e) for e in sorted(lookup, key = lambda x: x.value)])
    raise ZgyUserError("Value " + str(e) + " not accepted." +
                       " It should be one of " + valid + ".")

if __name__ == "__main__":
    help(ZgyReader)
    help(ZgyWriter)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



#!/usr/bin/env python3

print('Running' if __name__ == '__main__' else 'Importing', __file__)

import numpy as np
import sys
from time import time

# The new, pure Python API
from ..api import ZgyReader
# The old Python wrapper on top of C++ ZGY-Public
#from zgy import ZgyReader
from ..iterator import readall

def scanForRange(reader, verbose = False):
    realmin = []
    realmax = []
    begtime = time()
    for datastart, datasize, brick in readall(reader, dtype=np.float32):
        realmin.append(np.nanmin(brick))
        realmax.append(np.nanmax(brick))
    valuerange = (np.nanmin(realmin), np.nanmax(realmax))
    print("VALUE RANGE", valuerange)
    elapsed = time() - begtime
    voxels = np.product(reader.size) / (1024*1024)
    print("  {0:.1f} MVoxel read in {1:.1f} sec, {2:.1f} Mvoxel/s".format(
        voxels, elapsed, voxels/elapsed))
    return valuerange

def scanForHistogram(reader, valuerange, verbose = False):
    """
    Generate a histogram for all samples on the file.
    Note that if you want to verify the histogram on the file,
    the saved histogram range gives you the center of the first
    and last range but numpy wants the outer edges.
    """
    hh = None
    begtime = time()
    for datastart, datasize, brick in readall(reader, dtype=np.float32):
        h = np.histogram(brick, bins=256, range=valuerange)
        if hh is None:
            hh = h[0]
        else:
            hh += h[0]
    elapsed = time() - begtime
    voxels = np.product(reader.size) / (1024*1024)
    print("  {0:.1f} MVoxel read in {1:.1f} sec, {2:.1f} Mvoxel/s".format(
        voxels, elapsed, voxels/elapsed))
    return hh

def verifyHistogram(reader, verbose = False):
    h = reader.histogram
    binwidth = (h.max - h.min) / 255.0
    nprange = (h.min - binwidth/2, h.max + binwidth/2)
    hh = scanForHistogram(reader, nprange, verbose=verbose)
    stored = np.array(h.bin)
    delta = hh - stored
    #print("STORED", stored, sep="\n")
    #print("COMPUTED", hh, sep="\n")
    #print("COMPARED", delta, sep="\n")
    mismatch = np.array([(x, delta[x]) for x in range(len(delta)) if delta[x]])
    if len(mismatch):
        print("MISMATCH (bin, count) =", mismatch)
    else:
        print("HISTOGRAM CORRECT")

if __name__ == "__main__":

    np.seterr(all='raise')

    if len(sys.argv) <= 1:
        args = [ "/home/paal/git/Salmon/UnitTestData/Salmon/UnitTest/Salt2-v3.zgy" ]
    else:
        args = sys.argv[1:]

    for filename in args:
        with ZgyReader(filename) as reader:
            #reader.dump()
            print("size", reader.size, "datatype", reader.datatype)
            valuerange = scanForRange(reader, verbose=True)
            verifyHistogram(reader, verbose=True)

# Copyright 2017-2020, Schlumberger
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
